# Copyright (c) Alibaba, Inc. and its affiliates.
import os
import re
import numpy as np

from llmuses.benchmarks.data_adapter import DataAdapter
from llmuses.metrics.metrics import exact_match, weighted_mean
from llmuses.utils import normalize_score, jsonl_to_list
from llmuses.utils.logger import get_logger

# flake8: noqa

logger = get_logger()


DATASET_ID = 'modelscope/hellaswag'
SUBSET_LIST = ['default']


class HellaSwagAdapter(DataAdapter):

    choices = ['0', '1', '2', '3']

    def __init__(self,
                 subset_list: list = None,
                 metric_list: list = None,
                 few_shot_num: int = None,
                 train_split: str = 'train',
                 eval_split: str = 'validation',
                 **kwargs):

        if subset_list is None:
            subset_list = SUBSET_LIST

        if metric_list is None:
            metric_list = [{'name': 'WeightedAverageAccuracy', 'object': weighted_mean}]

        if few_shot_num is None:
            # Use 0-shot by default
            logger.info(f'Set 0-shot examples by system for HellaSwag.')
            few_shot_num = 0

        if few_shot_num != 0:
            logger.warning(f'few_shot_num should be 0 for HellaSwag, but got {few_shot_num}. Use 0-shot by default.')
            few_shot_num = 0

        super().__init__(subset_list=subset_list,
                         metric_list=metric_list,
                         few_shot_num=few_shot_num,
                         train_split=train_split,
                         eval_split=eval_split,
                         **kwargs)

    def load_from_disk(self, dataset_name_or_path, subset_list, work_dir, **kwargs) -> dict:
        data_dict = {}
        for subset_name in subset_list:
            data_dict[subset_name] = {}
            for split in [self.train_split, self.eval_split]:
                if os.path.exists(dataset_name_or_path):
                    file_path = os.path.join(dataset_name_or_path, f'hellaswag_{split}.jsonl')
                else:
                    file_path = os.path.join(work_dir, dataset_name_or_path, f'hellaswag_{split}.jsonl')
                if os.path.exists(file_path):
                    data_dict[subset_name][split] = jsonl_to_list(file_path)

        return data_dict

    def gen_prompt(self, input_d: dict, few_shot_list: list, **kwargs) -> dict:
        """
        Generate model prompt from raw data, unify the prompt format for HellaSwag benchmark.

        Args:
            input_d (dict): The raw input. A single data format of the HellaSwag:

            {
                'ind': 4,
                'activity_label': 'Removing ice from car',
                'ctx_a': 'Then, the man writes over the snow covering the window of a car, and a woman wearing winter clothes smiles.',
                'ctx_b': 'then',
                'ctx': 'Then, the man writes over the snow covering the window of a car, and a woman wearing winter clothes smiles. then',
                'endings': [', the man adds wax to the windshield and cuts it.', ', a person board a ski lift, while two men supporting the head of the person wearing winter clothes snow as the we girls sled.', ', the man puts on a christmas coat, knitted with netting.', ', the man continues removing the snow on his car.'],
                'source_id': 'activitynet~v_-1IBHYS3L-Y',
                'split': 'train',
                'split_type': 'indomain',
                'label': '3'
            }

        Returns:
            Refer to function: llmuses.benchmarks.data_adapter.DataAdapter.gen_prompt for details.
        """

        endings: list = [self._preprocess(ending) for ending in input_d['endings']]

        few_shot_prompts = [self._generate_prompt(input_d=sample, endings=endings, include_answer=True) for sample in few_shot_list]
        context: str = '\n'.join(few_shot_prompts) + '\n'
        context += self._generate_prompt(input_d=input_d, endings=endings, include_answer=False)

        ctx_continuation_pair_list = [(context.strip(), ' ' + cont.strip()) for cont in endings]

        return {'data': ctx_continuation_pair_list, 'multi_choices': self.choices}

    def get_gold_answer(self, input_d: dict) -> str:
        # Get the gold choice
        return input_d['label']

    def parse_pred_result(self, result: list, raw_input_d: dict = None, eval_type: str = 'checkpoint') -> str:
        """
        Parse the model output to get the answer. Could be the best choice index.

        Args:
            result: Predicted answer from the model. Usually a string for chat.
            raw_input_d: The raw input dict.
            eval_type: The evaluation type. e.g. checkpoint, service, custom.

        Returns:
            The parsed answer. Depending on the dataset. Usually a string for chat.
        """
        if eval_type == 'checkpoint':
            # answer: in the form of [-2.3, -4.5, ...], len of self.choices
            result = np.array(result)
            endings: list = [self._preprocess(ending) for ending in raw_input_d['endings']]
            completion_len = np.array([float(len(i)) for i in endings])
            best_choice_idx = np.argmax(result / completion_len)

            return str(best_choice_idx)
        elif eval_type == 'service':
            return result           # TODO: to be supported !
        elif eval_type == 'custom':
            return result           # TODO: to be supported !
        else:
            raise ValueError(f'Invalid eval_type: {eval_type}')

    def match(self, gold: str, pred: str) -> float:
        return exact_match(gold=str(gold), pred=str(pred))

    def compute_metric(self, review_res_list: list) -> float:
        """
        Compute evaluation result by specific metric.

        Args:
            review_res_list: review score list, e.g. [0, 1, 1, 0, ...]

        Returns:
            The metric score.
        """
        items = [(score, 1.0) for score in review_res_list]
        return weighted_mean(items)

    def gen_report(self, subset_score_map: dict, report_name: str = None) -> dict:
        """
        Generate the report for the model output.

        Args:
            subset_score_map: The subset-score mapping. e.g. {subset_name: (score, num), ...}
            report_name: The user-defined report name.

        Returns: A dict of metric calculation results. The format is like:
        {
            "name":"HellaSwag",
            "metric":"WeightedAverageAccuracy",
            "score":0.3389,
            "category":[
                {
                    "name":"DEFAULT",
                    "score":0.4128,
                    "subset":[
                        {
                            "name":"default",
                            "score":0.5632
                        },
                    ]
                }
            ],
            "total_num":7800
        }
        """
        total_num: int = sum([num for _, num in subset_score_map.values()])
        weighted_avg_acc: float = sum([score * num for score, num in subset_score_map.values()]) / total_num
        weighted_avg_acc = normalize_score(score=weighted_avg_acc)
        cate_avg_list = [{'name': subset_name, 'score': normalize_score(score=score)} for subset_name, (score, _) in subset_score_map.items()]

        category_d = dict(name='DEFAULT',
                          score=weighted_avg_acc,
                          subset=cate_avg_list)

        res_map = dict(name=report_name or 'hellaswag',
                       metric=self.metric_list[0]['name'],
                       score=weighted_avg_acc,
                       category=[category_d],
                       total_num=total_num)

        return res_map

    @classmethod
    def _preprocess(cls, text):
        text = text.strip()
        text = text.replace(' [title]', '. ')
        text = re.sub('\\[.*?\\]', '', text)
        text = text.replace('  ', ' ')
        return text

    @classmethod
    def _generate_prompt(cls, input_d: dict, endings: list, include_answer=True) -> str:
        """
        Generate prompt for HellaSwag dataset.

        Args:
            input_d: a single data of the hellaswag.
            endings:  preprocessed endings
            include_answer: bool

        Returns:

        """

        ctx = input_d['ctx_a'] + ' ' + input_d['ctx_b'].capitalize()
        example: str = cls._preprocess(input_d['activity_label'] + ': ' + ctx)

        if include_answer:
            example += '{}\n\n'.format(endings[int(input_d['label'])])

        return example



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) Allen Institute, and its affiliates.
# Copyright 2020 The HuggingFace Datasets Authors and the current dataset script contributor.

"""AI2 ARC (Abstraction and Reasoning Corpus) for General Artificial Intelligence Benchmark."""

"""AUTO GENERATED, DO NOT EDIT"""

import json
import os
import datasets

# flake8: noqa


_CITATION = """\
@article{allenai:arc,
      author    = {Peter Clark  and Isaac Cowhey and Oren Etzioni and Tushar Khot and
                    Ashish Sabharwal and Carissa Schoenick and Oyvind Tafjord},
      title     = {Think you have Solved Question Answering? Try ARC, the AI2 Reasoning Challenge},
      journal   = {arXiv:1803.05457v1},
      year      = {2018},
}
"""

_DESCRIPTION = """\
A new dataset of 7,787 genuine grade-school level, multiple-choice science questions, assembled to encourage research in
 advanced question-answering. The dataset is partitioned into a Challenge Set and an Easy Set, where the former contains
 only questions answered incorrectly by both a retrieval-based algorithm and a word co-occurrence algorithm. We are also
 including a corpus of over 14 million science sentences relevant to the task,
 and an implementation of three neural baseline models for this dataset. We pose ARC as a challenge to the community.

ARC-Easy:
    train: 2251
    test: 2376
    validation: 570

ARC-Challenge:
    train: 1119
    test: 1172
    validation: 299
"""

_URL = 'https://modelscope.oss-cn-beijing.aliyuncs.com/open_data/arc/ARC-V1-Feb2018.zip'

# tasks: ['ARC-Easy', 'ARC-Challenge']


class Ai2ArcConfig(datasets.BuilderConfig):
    """BuilderConfig for Ai2ARC."""

    def __init__(self, **kwargs):
        """BuilderConfig for Ai2Arc.

        Args:
          **kwargs: keyword arguments forwarded to super.
        """
        super(Ai2ArcConfig, self).__init__(version=datasets.Version('1.0.0', ''), **kwargs)


class Ai2Arc(datasets.GeneratorBasedBuilder):
    """
    The AI2 Reasoning Challenge (ARC) dataset.
    Subset: ARC-Easy, ARC-Challenge.
    """

    VERSION = datasets.Version('1.0.0')
    BUILDER_CONFIGS = [
        Ai2ArcConfig(
            name='ARC-Challenge',
            description="""\
          Challenge Set of 2590 “hard” questions (those that both a retrieval and a co-occurrence method fail to answer correctly)
          """,
        ),
        Ai2ArcConfig(
            name='ARC-Easy',
            description="""\
          Easy Set of 5197 questions
          """,
        ),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=_DESCRIPTION,
            # datasets.features.FeatureConnectors
            features=datasets.Features(
                {
                    'id': datasets.Value('string'),
                    'question': datasets.Value('string'),
                    'choices': datasets.features.Sequence(
                        {'text': datasets.Value('string'), 'label': datasets.Value('string')}
                    ),
                    'answerKey': datasets.Value('string')
                    # These are the features of your dataset like images, labels ...
                }
            ),
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage='https://allenai.org/data/arc',
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # dl_manager is a datasets.download.DownloadManager that can be used to
        # download and extract URLs
        dl_dir = dl_manager.download_and_extract(_URL)
        data_dir = os.path.join(dl_dir, 'ARC-V1-Feb2018-2')
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                # These kwargs will be passed to _generate_examples
                gen_kwargs={'filepath': os.path.join(data_dir, self.config.name, self.config.name + '-Train.jsonl')},
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                # These kwargs will be passed to _generate_examples
                gen_kwargs={'filepath': os.path.join(data_dir, self.config.name, self.config.name + '-Test.jsonl')},
            ),
            datasets.SplitGenerator(
                name=datasets.Split.VALIDATION,
                # These kwargs will be passed to _generate_examples
                gen_kwargs={'filepath': os.path.join(data_dir, self.config.name, self.config.name + '-Dev.jsonl')},
            ),
        ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        with open(filepath, encoding='utf-8') as f:
            for row in f:
                data = json.loads(row)
                answerkey = data['answerKey']
                id_ = data['id']
                question = data['question']['stem']
                choices = data['question']['choices']
                text_choices = [choice['text'] for choice in choices]
                label_choices = [choice['label'] for choice in choices]
                yield id_, {
                    'id': id_,
                    'answerKey': answerkey,
                    'question': question,
                    'choices': {'text': text_choices, 'label': label_choices},
                }



# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.ceval.ceval_adapter import CEVALAdapter, DATASET_ID, SUBSET_LIST, SUBJECT_MAPPING
from llmuses.benchmarks.ceval.ceval_adapter import CEVALAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa



# Copyright (c) Alibaba, Inc. and its affiliates.
import os, sys, time
from argparse import ArgumentParser
import subprocess


from llmuses.cli.base import CLICommand


current_path = os.path.dirname(os.path.abspath(__file__))
print(current_path)
root_path = os.path.dirname(current_path)
print(root_path)

def subparser_func(args):
    """ Function which will be called for a specific sub parser.
    """
    return PerfCMD(args)

def add_perf_args(parser):
    parser.add_argument(
        '--server-command', required=True, type=str, help='The start server command.')
    parser.add_argument(
        '--logdir', required=True, type=str, help='The monitor log save dir, tensorboard start at this path for display!')
    parser.add_argument(
        '--host', type=str, default='0.0.0.0', help='The tensorboard host'
    )    
    parser.add_argument(
        '--tensorboard-port', type=str, default='6006', help='The tensorboard port'
    )

def async_run_command_with_popen(cmd):
    sub_process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True,
        encoding='utf8')
    return sub_process

def start_monitor(args):
    cmd = ['python',
           '%s/perf/monitor.py'%root_path,
           '--logdir',
           args.logdir]
    print(cmd)
    p = async_run_command_with_popen(cmd)
    os.set_blocking(p.stdout.fileno(), False)
    return p

def start_tensorboard(args):
    cmd = ['tensorboard',
           '--logdir',
           args.logdir,
           '--host',
           args.host,
           '--port',
           args.tensorboard_port
           ]
    p = async_run_command_with_popen(cmd)
    os.set_blocking(p.stdout.fileno(), False)
    return p

def start_server(args):
    cmd = args.server_command
    print(cmd)
    sub_process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        shell=True,
        universal_newlines=True,
        encoding='utf8')

    os.set_blocking(sub_process.stdout.fileno(), False)
    return sub_process
 

def wait_for_workers(workers):
    while True:
        for idx, worker in enumerate(workers):
            if worker is None:
                continue
            # check worker is completed.
            if worker.poll() is None:
                for line in iter(worker.stdout.readline, ''):
                    if line != '':
                        sys.stdout.write(line)
                    else:
                        break
            else:
                print('Worker %s completed!'%idx)
                for line in iter(worker.stdout.readline, ''):
                    if line != '':
                        sys.stdout.write(line)
                    else:
                        break                
                workers[idx] = None

        is_all_completed = True
        for idx, worker in enumerate(workers):
            if worker is not None:
                is_all_completed = False
                break

        if is_all_completed:
            break
        time.sleep(0.1)
        
class PerfCMD(CLICommand):
    name = 'server'

    def __init__(self, args):
        self.args = args

    @staticmethod
    def define_args(parsers: ArgumentParser):
        """ define args for create pipeline template command.
        """
        parser = parsers.add_parser(PerfCMD.name)
        add_perf_args(parser)
        parser.set_defaults(func=subparser_func)

    def execute(self):
        # start monitor
        p_monitor = start_monitor(self.args)
        # start tensorboard
        p_tensorboard = start_tensorboard(self.args)        
        # start server
        p_server = start_server(self.args)
        
        wait_for_workers([p_monitor, p_tensorboard, p_server])
        
        
        
        


# Copyright (c) Alibaba, Inc. and its affiliates.
import os.path
from abc import ABC, abstractmethod
from typing import Any, Optional
import random

from llmuses.benchmarks import Benchmark
from llmuses.constants import DEFAULT_ROOT_CACHE_DIR, AnswerKeys
from llmuses.utils.logger import get_logger

logger = get_logger()


class DataAdapter(ABC):

    def __init__(self,
                 subset_list: list,
                 metric_list: list,
                 few_shot_num: Optional[int] = 0,
                 train_split: Optional[str] = None,
                 eval_split: Optional[str] = None,
                 prompt_template: str = '',
                 **kwargs):
        """
        Args:
            subset_list: list of subset names for the dataset.
            metric_list: list, the metric list to evaluate the model on specific benchmark.
            few_shot_num: int, number of few-shot examples. Default: 0
            train_split: str, usually for few-shot examples. e.g. 'train'
            eval_split: str, the target eval split name. e.g. 'test'
            prompt_template: str, the prompt template for the benchmark,
                e.g. for ARC, it is `The following are multiple choice questions, please output correct answer in the form of A or B or C or D, do not output explanation:`
        """
        self.subset_list = subset_list
        self.metric_list = metric_list
        self.few_shot_num = few_shot_num
        self.train_split = train_split
        self.eval_split = eval_split
        self.prompt_template = prompt_template
        self.config_kwargs = kwargs

    def load(self,
             dataset_name_or_path: str,
             subset_list: list = None,
             work_dir: Optional[str] = DEFAULT_ROOT_CACHE_DIR,
             datasets_hub: str = 'ModelScope',
             **kwargs) -> dict:
        """
        Load the dataset. Remote and local datasets are supported.
        You can rewrite this method to support your own local dataset, just follow the format of the output.

        Returns: {'subset_name': {'train': train_dataset, 'test': test_dataset}}
            train_dataset, test_dataset: Iterable dataset, object each item of which is a dict.

        """
        dataset_name_or_path = os.path.expanduser(dataset_name_or_path)
        if datasets_hub == 'Local':
            # Try to load dataset from local disk
            if os.path.isdir(dataset_name_or_path) and not os.path.exists(dataset_name_or_path):
                raise FileNotFoundError(f'Dataset path not found: {dataset_name_or_path}')

            logger.info(f'Loading dataset from local disk: >dataset_name: {dataset_name_or_path}  >work_dir: {work_dir}')
            data_dict = self.load_from_disk(dataset_name_or_path, subset_list, work_dir, **kwargs)
            if len(data_dict) == 0 or len(next(iter(data_dict.values()))) == 0:
                raise ValueError(f'Local dataset is empty: {dataset_name_or_path}')
        else:
            # Load dataset from remote
            logger.info(f'Loading dataset from {datasets_hub} hub: >dataset_name: {dataset_name_or_path}')
            data_dict = {}
            split_list = [split for split in [self.train_split, self.eval_split] if split is not None]
            if len(split_list) == 0:
                logger.error(f'Got empty split list: {split_list}')

            subset_list = subset_list if subset_list is not None else self.subset_list
            for sub_name in subset_list:
                data_dict[sub_name] = {}
                # e.g. train: few-shot, test: target dataset to evaluate
                for split in split_list:
                    dataset = Benchmark.load(dataset_name=dataset_name_or_path,
                                             subset=sub_name,
                                             split=split,
                                             hub=datasets_hub,
                                             work_dir=work_dir,
                                             **kwargs)

                    data_dict[sub_name].update({split: dataset})

        return data_dict

    def load_from_disk(self, *args, **kwargs) -> dict:
        """
        Load the dataset from local disk.
        If you want to support local dataset, please rewrite this method in xxx_data_adapter.
        """
        return {}

    def gen_prompts(self, data_dict: dict) -> dict:
        """
        Generate dataset prompts from raw input, unify the prompt format for different datasets.

        Args:
            data_dict:  Refer to the output of load method: llmuses.benchmarks.benchmark.Benchmark.load

        Returns:
            {'subset_name': [prompt_d_1, prompt_d_2, ...]}
            prompt_d_i (dict): refer to the output of gen_prompt method.

        e.g. train -- few-shot data, test -- target dataset to evaluate.
        """
        res_dict: dict = {}

        if self.few_shot_num and self.few_shot_num < 0:
            raise ValueError(f'Invalid shot_num: {self.few_shot_num} for few-shot evaluation.')

        logger.info(f'\n** Use default settings: \n'
                    f'>few_shot_num: {self.few_shot_num}, '
                    f'>few_shot_split: {self.train_split}, '
                    f'>target_eval_split: {self.eval_split}')

        for sub_name, sub_data_dict in data_dict.items():
            few_shot_data = []
            if self.few_shot_num and self.few_shot_num > 0:
                few_shot_random: bool = self.config_kwargs.get('few_shot_random', True)
                few_shot_data = self.get_fewshot_examples(
                    [item for item in sub_data_dict[self.train_split]],
                    self.few_shot_num,
                    few_shot_random=few_shot_random)

            res_dict[sub_name] = []
            for sample_d in sub_data_dict[self.eval_split]:
                prompt_d = self.gen_prompt(input_d=sample_d, subset_name=sub_name, few_shot_list=few_shot_data)
                prompt_d[AnswerKeys.RAW_INPUT] = sample_d
                res_dict[sub_name].append(prompt_d)

        rnd = random.Random()
        rnd.seed(42)
        for k, v in res_dict.items():
            rnd.shuffle(v)

        return res_dict

    @abstractmethod
    def gen_prompt(self, *args, **kwargs) -> Any:
        """
        Generate model prompt from raw input, unify the prompt format for different datasets.
        The input format is compatible with OpenAI Chat Completions APIs.
        Refer to: https://platform.openai.com/docs/guides/gpt/chat-completions-api

        Args:
            input_d (Any): The raw input. Depending on the dataset.

        Returns:
            For class MultiChoiceModelAdapter, the output format is:
                {'data': [full_prompt]},  -- full_prompt: str, the constructed prompt for each sample from dataset.

            For class ContinuationEvalModelAdapter, the output format is:
                {'data': ctx_continuation_pair_list, 'multi_choices': self.choices}
        """
        raise NotImplementedError

    @abstractmethod
    def get_gold_answer(self, input_d: Any) -> Any:
        """
        Parse the raw input labels (gold).

        Args:
            input_d: input raw data. Depending on the dataset.

        Returns:
            The parsed input. e.g. gold answer ... Depending on the dataset.
        """
        raise NotImplementedError

    @abstractmethod
    def parse_pred_result(self, result: Any, raw_input_d: dict = None, eval_type: str = 'checkpoint') -> Any:
        """
        Parse the predicted result and extract proper answer.

        Args:
            result: Predicted answer from the model. Usually a string for chat.
            raw_input_d: The raw input. Depending on the dataset.
            eval_type: 'checkpoint' or 'service' or `custom`, default: 'checkpoint'

        Returns:
            The parsed answer. Depending on the dataset. Usually a string for chat.
        """
        raise NotImplementedError

    @abstractmethod
    def match(self, gold: Any, pred: Any) -> Any:
        """
        Match the gold answer and the predicted answer.

        Args:
            gold (Any): The golden answer. Usually a string for chat/multiple-choice-questions.
                        e.g. 'A'
            pred (Any): The predicted answer. Usually a string for chat/multiple-choice-questions.
                        e.g. 'B'

        Returns:
            The match result. Usually a score (float) for chat/multiple-choice-questions.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_metric(self, review_res_list: list) -> Any:
        """
        Compute evaluation result by specific metrics.

        Args:
            review_res_list: list, the review result list, each item of which is match result for gold and pred.

        Attributes:
            DataAdapter.metric_func_map: metric_name -> metric_func mapping,
                e.g. {'WeightedAverageAccuracy': weighted_average_acc}

        Returns:
            Metric results.
        """
        raise NotImplementedError

    def gen_report(self, subset_score_map: dict, report_name: str = None) -> dict:
        """
        Generate report for the evaluation results for all subsets.

        Args:
            subset_score_map: The subset-score map.
                e.g. {subset_name: (score, num)}

            report_name: str, the user-defined report name. Default: None

        Returns: The evaluation report.  Note: should normalize the score by normalize_score method in utils.

        Here is a format example for ARC-Challenge:
        {
            "name":"ARC-Challenge",
            "metric":"WeightedAverageAccuracy",
            "score": 0.3389,
            "category":[
                {
                    "name":"DEFAULT",
                    "score": 0.3389,
                    "subset":[
                        {
                            "name":"ARC-Challenge",
                            "score": 0.3389
                        },
                    ]
                }
            ],
            "total_num":100
        }
        """
        raise NotImplementedError

    def get_fewshot_examples(self, data_list: list, k: int, few_shot_random: bool = True):

        if k > len(data_list):
            k = len(data_list)
        if few_shot_random:
            return random.sample(data_list, k)
        else:
            return data_list[:k]




from typing import Dict
import json

def get_query(model: str, prompt: str, **kwargs) -> Dict:
    """Get the query of from prompt and other parameters.

    Args:
        model (str): The model name.
        prompt (str): The input prompt.

    Returns:
        Dict: The request body.
    """
    messages = [{'role': 'user', 'content': prompt}]
    if prompt is None:
        raise Exception("Invalid parameters.")
    return {
        "model": model,
        "messages": messages,
        "stream": True,
        **kwargs
    }
    
def parse_responses(responses, **kwargs) -> Dict:
    """Parser responses and return number of request and response tokens.

    Args:
        responses (List[bytes]): List of http response body, for stream output,
            there are multiple responses, for general only one. 

    Returns:
        Dict: Return the prompt token and completion tokens.
    """
    last_response = responses[-1]
    js = json.loads(last_response)
    return js['usage']['prompt_tokens'], js['usage']['completion_tokens']


# Copyright (c) Alibaba, Inc. and its affiliates.
import json
import os
import glob
from typing import List, Union

from llmuses.config import TaskConfig
from llmuses.constants import OutputsStructure
from llmuses.tools.combine_reports import gen_table
from llmuses.utils import process_outputs_structure, yaml_to_dict
from llmuses.utils.logger import get_logger

logger = get_logger()


class Summarizer:

    @staticmethod
    def get_report(outputs_dir: str) -> List[dict]:
        res_list: list = []

        outputs_structure: dict = process_outputs_structure(outputs_dir, is_make=False)
        reports_dir: str = outputs_structure.get(OutputsStructure.REPORTS_DIR)
        if reports_dir is None:
            raise ValueError(f'No reports directory in {outputs_dir}')

        report_files: list = glob.glob(os.path.join(reports_dir, '*.json'))
        for report_file in report_files:
            with open(report_file, 'r') as f:
                res_list.append(json.load(f))

        report_table: str = gen_table([reports_dir])
        logger.info(f'*** Report table ***\n{report_table}')

        return res_list

    @staticmethod
    def get_report_from_cfg(task_cfg: Union[str, List[str], TaskConfig, List[TaskConfig]]) -> List[dict]:
        """
        Get report from cfg file.

        Args:
            task_cfg: task cfg file path. refer to llmuses/tasks/eval_qwen-7b-chat_v100.yaml

        Returns:
            list: list of report dict.
            A report dict is overall report on a benchmark for specific model.
        """
        candidate_task_cfgs: List[dict] = []

        if isinstance(task_cfg, str):
            task_cfg: dict = yaml_to_dict(task_cfg)
            candidate_task_cfgs = [task_cfg]
        elif isinstance(task_cfg, TaskConfig):
            task_cfg: dict = task_cfg.to_dict()
            candidate_task_cfgs = [task_cfg]
        elif isinstance(task_cfg, list):
            for task_cfg_item in task_cfg:
                if isinstance(task_cfg_item, str):
                    task_cfg_item: dict = yaml_to_dict(task_cfg_item)
                elif isinstance(task_cfg_item, TaskConfig):
                    task_cfg_item: dict = task_cfg_item.to_dict()
                candidate_task_cfgs.append(task_cfg_item)
        else:
            raise ValueError(f'Invalid task_cfg: {task_cfg}')

        final_res_list: list = []
        outputs_dir_list: list = []
        for candidate_task in candidate_task_cfgs:
            logger.info(f'**Task cfg: {candidate_task}')
            outputs_dir: str = candidate_task.get('outputs')
            outputs_dir: str = os.path.expanduser(outputs_dir)
            if outputs_dir is None:
                raise ValueError(f'No outputs_dir in {task_cfg}')
            outputs_dir_list.append(outputs_dir)
        outputs_dir_list = list(set(outputs_dir_list))

        for outputs_dir_item in outputs_dir_list:
            res_list: list = Summarizer.get_report(outputs_dir=outputs_dir_item)
            final_res_list.extend(res_list)

        return final_res_list


if __name__ == '__main__':
    cfg_file = 'registry/tasks/eval_qwen-7b-chat_v100.yaml'
    report_list = Summarizer.get_report_from_cfg(cfg_file)

    print(report_list)



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) EleutherAI, Inc. and its affiliates.
# flake8: noqa
import os
import sys
from typing import List, Any, Union, Dict
import numpy as np
import time
from abc import ABC, abstractmethod
from copy import deepcopy

import torch
from torch import dtype

from llmuses.constants import DEFAULT_ROOT_CACHE_DIR
from llmuses.models.custom import CustomModel
from llmuses.models.template import get_template, StopWordsCriteria, MODEL_TEMPLATE_MAP
from llmuses.utils.logger import get_logger
from transformers import StoppingCriteriaList

logger = get_logger()

# Notes:
# - modelscope>=1.9.5


def get_model_cache_dir(root_cache_dir: str):
    model_cache_dir = os.path.join(root_cache_dir, 'models')
    model_cache_dir = os.path.expanduser(model_cache_dir)
    os.makedirs(model_cache_dir, exist_ok=True)
    return model_cache_dir


class BaseModelAdapter(ABC):
    """
    Base class for model adapter.
    """

    def __init__(self, model, tokenizer, model_cfg: dict):
        """
        Args:
            model: The model instance which is compatible with
                AutoModel/AutoModelForCausalLM/AutoModelForSeq2SeqLM of transformers.
            tokenizer: The tokenizer instance which is compatible with AutoTokenizer of transformers.
            model_cfg:
                Attributes: model_id, model_revision, device_map, torch_dtype
        """
        self.model = model
        self.tokenizer = tokenizer
        self.model_cfg = model_cfg

    @abstractmethod
    @torch.no_grad()
    def predict(self, *args, **kwargs) -> Any:
        """
        Model prediction func.
        """
        raise NotImplementedError


class MultiChoiceModelAdapter(BaseModelAdapter):
    """ The multi-choice model adapter. """

    _DEFAULT_MAX_LENGTH = 2048

    def __init__(self,
                 model_id: str,
                 device_map: str = 'auto',
                 torch_dtype: dtype = torch.bfloat16,
                 model_revision: str = None,
                 max_length: int = None,
                 cache_dir: str = DEFAULT_ROOT_CACHE_DIR,
                 **kwargs):
        """
        Args:
            model_id: The model id on ModelScope, or local model_dir.  TODO: torch.nn.module to be supported.
            device_map: The device map for model inference.
            torch_dtype: The torch dtype for model inference. Default: torch.bfloat16.
            model_revision: The model revision on ModelScope. Default: None.
            max_length: The max length of input sequence. Default: None.
            **kwargs: Other args.
        """
        model_cache_dir = get_model_cache_dir(cache_dir)

        self.model_id: str = model_id
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        logger.warning(f'**Device: {self.device}')

        torch_dtype = torch_dtype if torch_dtype is not None else 'auto'

        model_cfg: dict = dict()
        model_cfg['model_id'] = model_id
        model_cfg['device_map'] = device_map
        model_cfg['torch_dtype'] = str(torch_dtype)

        from modelscope.utils.hf_util import AutoModelForCausalLM, AutoTokenizer
        from modelscope import snapshot_download

        try:
            model_dir = snapshot_download(self.model_id, cache_dir=model_cache_dir, local_files_only=True)
            logger.warning('**Use local_files_only to load model **')
        except:
            model_dir = snapshot_download(self.model_id,
                                          revision=model_revision,
                                          cache_dir=model_cache_dir, )
            logger.warning('**Load model from ModelScope hub **')

        tokenizer = AutoTokenizer.from_pretrained(model_dir,    # self.model_id
                                                  revision=model_revision,
                                                  trust_remote_code=True,)

        model = AutoModelForCausalLM.from_pretrained(model_dir,  # self.model_id
                                                     revision=model_revision,
                                                     device_map=device_map,
                                                     trust_remote_code=True,
                                                     torch_dtype=torch_dtype,)

        # model.generation_config = GenerationConfig.from_pretrained(model_id, trust_remote_code=True)

        super().__init__(model=model, tokenizer=tokenizer, model_cfg=model_cfg)

        self._max_length = max_length

    @property
    def max_length(self):
        if self._max_length:
            return self._max_length
        seqlen_config_attrs = ('n_positions', 'max_position_embeddings', 'n_ctx')
        for attr in seqlen_config_attrs:
            if hasattr(self.model.config, attr):
                return getattr(self.model.config, attr)
        if hasattr(self.tokenizer, 'model_max_length'):
            if self.tokenizer.model_max_length == 1000000000000000019884624838656:
                return self._DEFAULT_MAX_LENGTH
            return self.tokenizer.model_max_length
        return self._DEFAULT_MAX_LENGTH

    @torch.no_grad()
    def predict(self, inputs: dict, infer_cfg: dict = None) -> dict:
        """
        Multi-choice model prediction func.

        Args:
            inputs (dict): The inputs for a doc. Format:
                {'data': [full_prompt], 'multi_choices': ['A', 'B', 'C', 'D']}

            infer_cfg (dict): inference configuration.

        Returns:
            res (dict): The model prediction results. Format:
            {
              'choices': [
                {
                  'index': 0,
                  'message': {
                    'content': [-14.9609, -13.6015, ...],  # loglikelihood values for inputs context-continuation pairs.
                    'role': 'assistant'
                  }
                }
              ],
              'created': 1677664795,
              # For models on the ModelScope or HuggingFace, concat model_id and revision with "-".
              'model': 'gpt-3.5-turbo-0613',
              'object': 'chat.completion',
              'usage': {
                'completion_tokens': 17,
                'prompt_tokens': 57,
                'total_tokens': 74
              }
            }
        """
        infer_cfg = infer_cfg or {}
        self.model.generation_config.update(**infer_cfg)

        input_data = inputs['data']
        multi_choices = inputs['multi_choices']

        output, input_info = self._get_logits(self.tokenizer, self.model, input_data)
        assert output.shape[0] == 1
        logits = output.flatten()

        choice_logits = [logits[self.tokenizer(ch)['input_ids'][-1:]] for ch in multi_choices]
        softval = torch.nn.functional.softmax(torch.tensor(choice_logits).float(), dim=0)

        if softval.dtype in {torch.bfloat16, torch.float16}:
            softval = softval.to(dtype=torch.float32)
        probs = softval.detach().cpu().numpy()
        pred: str = multi_choices[int(np.argmax(probs))]        # Format: A or B or C or D

        res_d = {
            'choices': [
                {
                    'index': 0,
                    'message': {
                        'content': pred,
                        'role': 'assistant'
                    }
                }
            ],
            'created': time.time(),
            'model': self.model_id,
            'object': 'chat.completion',
            'usage': {}
        }

        return res_d

    @staticmethod
    def _get_logits(tokenizer, model, inputs: List[str]):
        input_ids = tokenizer(inputs, padding=False)['input_ids']
        input_ids = torch.tensor(input_ids, device=model.device)
        tokens = {'input_ids': input_ids}

        outputs = model(input_ids)['logits']
        logits = outputs[:, -1, :]
        log_probs = torch.nn.functional.softmax(logits, dim=-1)
        return log_probs, {'tokens': tokens}


class ContinuationLogitsModelAdapter(MultiChoiceModelAdapter):

    def __init__(self,
                 model_id: str,
                 device_map: str = 'auto',
                 torch_dtype: dtype = torch.bfloat16,
                 model_revision: str = None,
                 cache_dir: str = DEFAULT_ROOT_CACHE_DIR,
                 **kwargs):
        """
        Continuation-logits model adapter.

        Args:
            model_id: The model id on ModelScope, or local model_dir.
            device_map: The device map for model inference.
            torch_dtype: The torch dtype for model inference. Default: torch.bfloat16.
            model_revision: The model revision on ModelScope. Default: None.
            **kwargs: Other args.
        """

        super().__init__(model_id=model_id,
                         device_map=device_map,
                         torch_dtype=torch_dtype,
                         model_revision=model_revision,
                         cache_dir=cache_dir,
                         **kwargs)

    @torch.no_grad()
    def predict(self, inputs: dict, infer_cfg: dict = None) -> dict:
        """
        Multi-choice model prediction func.
        Args:
            inputs (dict): The inputs for a doc. Format:
                {'data': [(context, continuation), ...]}
            infer_cfg (dict): inference configuration.
        Returns:
            res (dict): The model prediction results. Format:
            {
              'choices': [
                {
                  'index': 0,
                  'message': {
                    'content': [-14.9609, -13.6015, ...],  # loglikelihood values for inputs context-continuation pairs.
                    'role': 'assistant'
                  }
                }
              ],
              'created': 1677664795,
              # For models on the ModelScope or HuggingFace, concat model_id and revision with "-".
              'model': 'gpt-3.5-turbo-0613',
              'object': 'chat.completion',
              'usage': {
                'completion_tokens': 17,
                'prompt_tokens': 57,
                'total_tokens': 74
              }
            }
        """
        infer_cfg = infer_cfg or {}

        pred_list: list = self.loglikelihood(inputs=inputs['data'], infer_cfg=infer_cfg)

        res_d = {
            'choices': [
                {
                    'index': 0,
                    'message': {
                        'content': pred_list,
                        'role': 'assistant'
                    }
                }
            ],
            'created': time.time(),
            'model': self.model_id,
            'object': 'chat.completion',
            'usage': {}
        }
        return res_d

    def loglikelihood(self, inputs: list, infer_cfg: dict = None) -> list:
        self.model.generation_config.update(**infer_cfg)
        # To predict one doc
        doc_ele_pred = []
        for ctx, continuation in inputs:

            # ctx_enc shape: [context_tok_len]  cont_enc shape: [continuation_tok_len]
            ctx_enc, cont_enc = self._encode_pair(ctx, continuation)

            inputs_tokens = torch.tensor(
                (ctx_enc.tolist() + cont_enc.tolist())[-(self.max_length + 1):][:-1],
                dtype=torch.long,
                device=self.model.device).unsqueeze(0)

            logits = self.model(inputs_tokens)[0]
            logits = torch.nn.functional.log_softmax(logits.float(), dim=-1)

            logits = logits[:, -len(cont_enc):, :]
            cont_enc = cont_enc.unsqueeze(0).unsqueeze(-1)
            logits = torch.gather(logits.cpu(), 2, cont_enc.cpu()).squeeze(-1)

            choice_score = float(logits.sum())
            doc_ele_pred.append(choice_score)

        # e.g. [-2.3, -9.2, -12.9, 1.1], length=len(choices)
        return doc_ele_pred

    def _encode_pair(self, context, continuation):
        n_spaces = len(context) - len(context.rstrip())
        if n_spaces > 0:
            continuation = context[-n_spaces:] + continuation
            context = context[:-n_spaces]

        whole_enc = self.tokenizer(context + continuation, padding=False)['input_ids']
        whole_enc = torch.tensor(whole_enc, device=self.device)

        context_enc = self.tokenizer(context, padding=False)['input_ids']
        context_enc = torch.tensor(context_enc, device=self.device)

        context_enc_len = len(context_enc)
        continuation_enc = whole_enc[context_enc_len:]

        return context_enc, continuation_enc


class ChatGenerationModelAdapter(BaseModelAdapter):

    def __init__(self,
                 model_id: str,
                 model_revision: str,
                 device_map: str = 'auto',
                 torch_dtype: dtype = torch.float16,
                 cache_dir: str = DEFAULT_ROOT_CACHE_DIR,
                 **kwargs):
        """
        Chat completion model adapter. Tasks of chat and generation are supported.

        Args:
            model_id: The model id on ModelScope, or local model_dir.
            model_revision: The model revision on ModelScope. Default: None.
            device_map: The device map for model inference.
            torch_dtype: The torch dtype for model inference. Default: torch.float16.
            **kwargs: Other args.
        """
        model_cache_dir = get_model_cache_dir(root_cache_dir=cache_dir)

        self.model_id: str = model_id
        self.model_revision: str = model_revision
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        logger.warning(f'**Device: {self.device}')

        torch_dtype = torch_dtype if torch_dtype is not None else 'auto'

        model_cfg: dict = dict()
        model_cfg['model_id'] = model_id
        model_cfg['device_map'] = device_map
        model_cfg['torch_dtype'] = str(torch_dtype)

        from modelscope.utils.hf_util import AutoModelForCausalLM, AutoTokenizer
        from modelscope import snapshot_download

        try:
            model_dir = snapshot_download(self.model_id, cache_dir=model_cache_dir, local_files_only=True)
            logger.warning('**Use local_files_only to load model **')
        except:
            model_dir = snapshot_download(self.model_id,
                                          revision=model_revision,
                                          cache_dir=model_cache_dir, )
            logger.warning('**Load model from ModelScope hub **')

        tokenizer = AutoTokenizer.from_pretrained(model_dir,
                                                  revision=model_revision,
                                                  trust_remote_code=True,)

        model = AutoModelForCausalLM.from_pretrained(model_dir,
                                                     revision=model_revision,
                                                     device_map=device_map,
                                                     trust_remote_code=True,
                                                     torch_dtype=torch_dtype, )

        self.origin_tokenizer = deepcopy(tokenizer)

        self.generation_config, self.generation_template = self._parse_generation_config(tokenizer, model)
        logger.info(f'**Generation config init: {self.generation_config.to_dict()}')

        super().__init__(model=model, tokenizer=self.generation_template.tokenizer, model_cfg=model_cfg)

    def _parse_generation_config(self, tokenizer, model):
        from modelscope.utils.hf_util import GenerationConfig

        generation_config = getattr(model, 'generation_config', GenerationConfig())

        try:
            remote_config = GenerationConfig.from_pretrained(
                self.model_id,
                revision=self.model_revision,
                trust_remote_code=True)
            generation_config.update(**remote_config.to_dict())
        except:
            logger.warning(f'Failed to get generation config of {self.model_id} from model hub, use default.')

        # Parse templates for chat-completion
        if os.path.exists(self.model_id):
            logger.warning(f'Got local model: {self.model_id}, '
                           f'please make sure the type of path in the form of `/path/to/your_model_name`')
        model_name = os.path.basename(os.path.normpath(self.model_id))      # TODO: check compatibility with path
        logger.info(f'**Model name: {model_name}')
        template_type: str = MODEL_TEMPLATE_MAP.get(model_name, [None, None])[0]
        if template_type is None:
            from llmuses.models.template import TemplateType
            template_type = TemplateType.default_generation
            logger.warning(f'Failed to get template type of {self.model_id}, use default: {template_type}')

        logger.info(f'**Template type of generation: {template_type}')
        generation_template = get_template(model_name, template_type, tokenizer)

        if tokenizer.eos_token_id is not None:
            generation_config.eos_token_id = tokenizer.eos_token_id
        if tokenizer.pad_token_id is not None:
            generation_config.pad_token_id = tokenizer.pad_token_id
        if generation_config.max_new_tokens is not None:
            generation_config.max_length = 20

        return generation_config, generation_template

    def _model_generate(self, query: str, infer_cfg: dict) -> str:
        example = dict(query=query,
                       history=[],
                       system=None)

        inputs = self.generation_template.encode(example)
        input_ids = inputs['input_ids']
        input_ids = torch.tensor(input_ids)[None].to(self.device)
        attention_mask = torch.ones_like(input_ids).to(self.device)

        # Process infer_cfg
        infer_cfg = infer_cfg or {}
        if isinstance(infer_cfg.get('num_return_sequences'), int) and infer_cfg['num_return_sequences'] > 1:
            infer_cfg['do_sample'] = True

        # TODO: stop settings
        stop = infer_cfg.get('stop', None)
        eos_token_id = self.tokenizer.encode(stop, add_special_tokens=False)[0] \
            if stop else self.tokenizer.eos_token_id

        if eos_token_id is not None:
            infer_cfg['eos_token_id'] = eos_token_id
            infer_cfg['pad_token_id'] = eos_token_id  # setting eos_token_id as pad token


        self.generation_config.update(**infer_cfg)

        # stopping
        stop_words = [self.generation_template.suffix[-1]]
        decode_kwargs = {}
        stopping_criteria = StoppingCriteriaList(
            [StopWordsCriteria(self.tokenizer, stop_words, **decode_kwargs)])

        # Run inference
        output_ids = self.model.generate(
            input_ids=input_ids,
            attention_mask=attention_mask,
            generation_config=self.generation_config,
            stopping_criteria=stopping_criteria, )

        response = self.tokenizer.decode(output_ids[0, len(input_ids[0]):], True, **decode_kwargs)
        return response

    @torch.no_grad()
    def predict(self, inputs: Union[str, dict, list], infer_cfg: dict = dict({})) -> dict:

        # Process inputs
        if isinstance(inputs, str):
            query = inputs
        elif isinstance(inputs, dict):
            query = inputs['data'][0]
        elif isinstance(inputs, list):
            query = '\n'.join(inputs)
        else:
            raise TypeError(f'Unsupported inputs type: {type(inputs)}')

        response = self._model_generate(query, infer_cfg)

        choices_list = [
            {'index': 0,
             'message': {'content': response,
                         'role': 'assistant'}
             }
        ]

        res_d = {
            'choices': choices_list,
            'created': time.time(),
            'model': self.model_id,
            'object': 'chat.completion',
            'usage': {}
        }

        return res_d


class CustomModelAdapter(BaseModelAdapter):

    def __init__(self, custom_model: CustomModel, **kwargs):
        """
        Custom model adapter.

        Args:
            custom_model: The custom model instance.
            **kwargs: Other args.
        """
        self.custom_model = custom_model
        super(CustomModelAdapter, self).__init__(model=None, tokenizer=None, model_cfg=custom_model.config)

    def predict(self, inputs: Union[str, dict, list], **kwargs) -> Dict[str, Any]:
        """
        Model prediction func.

        Args:
            inputs (Union[str, dict, list]): The input data. Depending on the specific model.
                str: 'xxx'
                dict: {'data': [full_prompt]}
                list: ['xxx', 'yyy', 'zzz']
            **kwargs: kwargs

        Returns:
            res (dict): The model prediction results. Format:
            {
              'choices': [
                {
                  'index': 0,
                  'message': {
                    'content': 'xxx',
                    'role': 'assistant'
                  }
                }
              ],
              'created': 1677664795,
              'model': 'gpt-3.5-turbo-0613',   # should be model_id
              'object': 'chat.completion',
              'usage': {
                'completion_tokens': 17,
                'prompt_tokens': 57,
                'total_tokens': 74
              }
            }
        """

        # Process inputs
        if isinstance(inputs, str):
            query = inputs
        elif isinstance(inputs, dict):
            query = inputs['data'][0]
        elif isinstance(inputs, list):
            query = '\n'.join(inputs)
        else:
            raise TypeError(f'Unsupported inputs type: {type(inputs)}')

        return self.custom_model.predict(prompt=query, **kwargs)




# Copyright (c) Alibaba, Inc. and its affiliates.
# flake8: noqa

import ast
import re


# from . import utils as ann_utils
from llmuses.constants import ArenaWinner
from llmuses.utils.logger import get_logger

logger = get_logger()

one_score_pattern = re.compile('\[\[(\d+\.?\d*)\]\]')
one_score_pattern_backup = re.compile('\[(\d+\.?\d*)\]')


# modified from: https://github.com/lm-sys/FastChat/blob/main/fastchat/eval/eval_gpt_review.py#L47
# does not work with batched completions
def lmsys_parser(completion, output_format):
    if output_format == '[[rating]]':
        match = re.search(one_score_pattern, completion)
        if not match:
            match = re.search(one_score_pattern_backup, completion)

        if match:
            rating = ast.literal_eval(match.groups()[0])
        else:
            logger.error(f'Content: {completion}\n'
                         'You must manually fix the score.')
            rating = -1

        return rating
    if output_format == '[[rating_a,rating_b]]':
        try:
            score_pair = completion.split('\n')[0]
            score_pair = score_pair.replace(',', ' ')
            sp = score_pair.split(' ')
            if len(sp) == 2:
                score_1 = float(sp[0])
                score_2 = float(sp[1])
                if score_1 > score_2:
                    winner = ArenaWinner.MODEL_A
                elif score_1 < score_2:
                    winner = ArenaWinner.MODEL_B
                else:
                    if score_1 == score_1 == -1:
                        winner = ArenaWinner.UNKNOWN
                    winner = ArenaWinner.TIE
                return winner, [score_1, score_2]
            else:
                raise Exception('Invalid score pair.')
        except Exception as e:
            logger.error(
                f'{e}\nContent: {completion}\nYou must manually fix the score pair.'
            )
            return ArenaWinner.UNKNOWN, [-1, -1]
    elif output_format == '[[A]]':
        if '[[A]]' in completion:
            winner = ArenaWinner.MODEL_A
        elif '[[B]]' in completion:
            winner = ArenaWinner.MODEL_B
        elif '[[C]]' in completion:
            winner = ArenaWinner.TIE
        else:
            logger.error(
                f'\nContent: {completion}\nYou must manually fix the score.')
            winner = ArenaWinner.UNKNOWN
        return winner


def ranking_parser(completion, **kwargs):
    try:
        if isinstance(completion, str):
            ordered_completions = ast.literal_eval(completion)
        else:
            ordered_completions = completion

        rank = [c for c in ordered_completions
                if c['model'] == 'model_a'][0]['rank']
        assert rank in [1, 2]

        return ArenaWinner.MODEL_A if rank == 1 else ArenaWinner.MODEL_B
    except Exception as e:
        logger.error(f'{e}\nContent: {completion}\n'
                     'You must manually fix the score pair.')
        return ArenaWinner.UNKNOWN



# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.arc.arc_adapter import ARCAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.arc.arc_adapter import ARCAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa



# Copyright (c) Alibaba, Inc. and its affiliates.

import os
from typing import Union

import cachetools
from cachetools import Cache as CachetoolsCache
from pympler import asizeof
from datetime import datetime, timedelta
import pickle

from llmuses.constants import DEFAULT_ROOT_CACHE_DIR
from llmuses.utils.logger import get_logger

logger = get_logger()


DEFAULT_CACHE_MAXSIZE = 1 * 1024 * 1024 * 1024  # 1 GB
DEFAULT_CACHE_EXPIRE = 60 * 60 * 24             # 1 day (seconds)
DEFAULT_MEM_CACHE_PATH = os.environ.get('MEM_CACHE_PATH',
                                        os.path.join(os.path.expanduser(DEFAULT_ROOT_CACHE_DIR),
                                                     'mem_cache', 'global_cache.pkl'))


class Cache:

    # TODO:   by xingjun.wxj@alibaba-inc.com
    #  1. atomic operation for saving cache
    #  2. consider the distributed env

    @classmethod
    def lru_cache(cls, maxsize: int = DEFAULT_CACHE_MAXSIZE):
        return cachetools.LRUCache(maxsize=maxsize, getsizeof=asizeof.asizeof)

    @classmethod
    def ttl_cache(cls, max_size: float = DEFAULT_CACHE_MAXSIZE, expire: float = DEFAULT_CACHE_EXPIRE):
        return cachetools.TTLCache(maxsize=max_size,
                                   ttl=timedelta(seconds=expire),
                                   timer=datetime.now,
                                   getsizeof=asizeof.asizeof)

    @classmethod
    def load(cls, path: str) -> Union[CachetoolsCache, None]:
        """
        Load cache from disk. Pickle is used for serialization.

        Args:
            path: The local path to load the cache.

        Returns:
            The cache instance loaded from disk. Should be cachetools.Cache or None.
        """
        if os.path.exists(path):
            logger.info(f'** Loading cache from {path} ...')
            with open(path, 'rb') as f:
                return pickle.load(f)
        else:
            return None

    @classmethod
    def save(cls, cache: CachetoolsCache, path: str = DEFAULT_MEM_CACHE_PATH):
        """
        Dump memory cache to disk. Pickle is used for serialization.

        Args:
            cache: The cache instance to be saved.
            path: The local path to save the cache.

        Returns: None
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as f:
            pickle.dump(cache, f)
        logger.info(f'** Cache saved to {path} !')


def init_mem_cache(method: str = 'ttl', cache_file_path: str = DEFAULT_MEM_CACHE_PATH) -> CachetoolsCache:
    """
    Initialize memory cache.

    Args:
        method (str): 'ttl' or 'lru', see https://cachetools.readthedocs.io/en/latest/ for details.
        cache_file_path (str): The local cache path. Should be a pickle file.

    Returns:
        The cache instance. Should be cachetools.Cache.
    """
    logger.info(f'** Initializing memory cache with method `{method}` ... \n')
    mem_cache = Cache.load(path=cache_file_path)
    if mem_cache is None:
        if method == 'ttl':
            mem_cache = Cache.ttl_cache()
        elif method == 'lru':
            mem_cache = Cache.lru_cache()
        else:
            raise ValueError(f'Unknown cache method {method}. Please choose from `ttl` or `lru`.')

    return mem_cache



# Copyright (c) Alibaba, Inc. and its affiliates.

"""Mathematics Aptitude Test of Heuristics (MATH) dataset."""

import json
import os

import datasets


_CITATION = """\
@article{hendrycksmath2021,
  title={Measuring Mathematical Problem Solving With the MATH Dataset},
  author={Dan Hendrycks
    and Collin Burns
    and Saurav Kadavath
    and Akul Arora
    and Steven Basart
    and Eric Tang
    and Dawn Song
    and Jacob Steinhardt},
  journal={arXiv preprint arXiv:2103.03874},
  year={2021}
}
"""


_DESCRIPTION = """\
The Mathematics Aptitude Test of Heuristics (MATH) dataset consists of problems
from mathematics competitions, including the AMC 10, AMC 12, AIME, and more.
Each problem in MATH has a full step-by-step solution, which can be used to teach
models to generate answer derivations and explanations.
"""


_HOMEPAGE = 'https://github.com/hendrycks/math'


_LICENSE = 'https://github.com/hendrycks/math/blob/main/LICENSE'


# Original data URL: "https://people.eecs.berkeley.edu/~hendrycks/MATH.tar"
_URL = 'https://sail-moe.oss-cn-hangzhou.aliyuncs.com/open_data/math/MATH.zip'


class CompetitionMathDataset(datasets.GeneratorBasedBuilder):
    """Mathematics Aptitude Test of Heuristics (MATH) dataset."""

    VERSION = datasets.Version('1.0.0')

    def _info(self):
        features = datasets.Features(
            {
                'problem': datasets.Value('string'),
                'level': datasets.Value('string'),
                'type': datasets.Value('string'),
                'solution': datasets.Value('string'),
            }
        )
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=features,
            supervised_keys=None,
            homepage=_HOMEPAGE,
            license=_LICENSE,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        download_dir = dl_manager.download_and_extract(_URL)
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={'data_dir': dl_manager.iter_files(os.path.join(download_dir, 'MATH', 'train'))},
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                gen_kwargs={'data_dir': dl_manager.iter_files(os.path.join(download_dir, 'MATH', 'test'))},
            ),
        ]

    def _generate_examples(self, data_dir):
        """Yields examples as (key, example) tuples."""
        for id_, filepath in enumerate(data_dir):
            with open(filepath, 'rb') as fin:
                example = json.load(fin)
                yield id_, example



# Copyright (c) Alibaba, Inc. and its affiliates.

# flake8: noqa


DATASET_ID = 'modelscope/humaneval'
SUBSET_LIST = ['openai_humaneval']

# Note: ONLY FOR CLASS IMPORT, No implementation here.

# Example:
# {"task_id": "HumanEval/0", "prompt": "from typing import List\n\n\ndef has_close_elements(numbers: List[float], threshold: float) -> bool:\n    \"\"\" Check if in given list of numbers, are any two numbers closer to each other than\n    given threshold.\n    >>> has_close_elements([1.0, 2.0, 3.0], 0.5)\n    False\n    >>> has_close_elements([1.0, 2.8, 3.0, 4.0, 5.0, 2.0], 0.3)\n    True\n    \"\"\"\n", "entry_point": "has_close_elements", "canonical_solution": "    for idx, elem in enumerate(numbers):\n        for idx2, elem2 in enumerate(numbers):\n            if idx != idx2:\n                distance = abs(elem - elem2)\n                if distance < threshold:\n                    return True\n\n    return False\n", "test": "\n\nMETADATA = {\n    'author': 'jt',\n    'dataset': 'test'\n}\n\n\ndef check(candidate):\n    assert candidate([1.0, 2.0, 3.9, 4.0, 5.0, 2.2], 0.3) == True\n    assert candidate([1.0, 2.0, 3.9, 4.0, 5.0, 2.2], 0.05) == False\n    assert candidate([1.0, 2.0, 5.9, 4.0, 5.0], 0.95) == True\n    assert candidate([1.0, 2.0, 5.9, 4.0, 5.0], 0.8) == False\n    assert candidate([1.0, 2.0, 3.0, 4.0, 5.0, 2.0], 0.1) == True\n    assert candidate([1.1, 2.2, 3.1, 4.1, 5.1], 1.0) == True\n    assert candidate([1.1, 2.2, 3.1, 4.1, 5.1], 0.5) == False\n\n"}


class HumanevalAdapter:
    """
    A placeholder for humaneval adapter, see HumanevalEvaluator for implementation.
    """

    def __init__(self):
        ...



import logging
import os


logger = logging.getLogger('perf')


def enable_logging():
    level = os.environ.get('LOGGING_LEVEL', 'info')
    if level is not None:  # set logging level.
        if level not in ['info', 'debug']:
            # set logging level env, but invalid value, use default.
            level = 'info'
        if level == 'info':
            logger.setLevel(logging.INFO)
        else:
            logger.setLevel(logging.DEBUG)
        # set default logging handler
        console_handler = logging.StreamHandler()
        #formatter = logging.Formatter(
        #    '%(asctime)s - %(name)s - %(filename)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s'  # noqa E501
        #)
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'  # noqa E501
        )
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)


# in release disable dashscope log
# you can enable dashscope log for debugger.
enable_logging()



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) EleutherAI, Inc. and its affiliates.
import glob
import json
import os

from llmuses.benchmarks import DataAdapter
from llmuses.metrics.metrics import weighted_mean
from llmuses.utils import normalize_score
from llmuses.utils.logger import get_logger
# flake8: noqa

logger = get_logger()

DATASET_ID = 'modelscope/competition_math'
SUBSET_LIST = ['default']


class CompetitionMathAdapter(DataAdapter):
    """ TODO: To be tested for all models. """

    def __init__(self,
                 subset_list: list = None,
                 metric_list: list = None,
                 few_shot_num: int = None,
                 train_split: str = 'train',
                 eval_split: str = 'test',
                 **kwargs):

        if subset_list is None:
            subset_list = SUBSET_LIST

        if metric_list is None:
            metric_list = [{'name': 'WeightedAverageAccuracy', 'object': weighted_mean}]

        if few_shot_num is None:
            # Use 4-shot by default
            logger.info(f'Set 4-shot examples by system for MATH.')
            few_shot_num = 4

        if few_shot_num != 4 and few_shot_num != 0:
            logger.error(f'The MATH benchmark ONLY supports 4-shot by system or 0-shot settings, '
                         f'but got {self.few_shot_num}. Use 4-shot by default.')
            few_shot_num = 4

        super().__init__(subset_list=subset_list,
                         metric_list=metric_list,
                         few_shot_num=few_shot_num,
                         train_split=train_split,
                         eval_split=eval_split,
                         **kwargs)

    def load_from_disk(self, dataset_name_or_path, subset_list, work_dir, **kwargs) -> dict:
        data_dict: dict = {}
        for subset_name in subset_list:
            for split_name in [self.train_split, self.eval_split]:
                if os.path.exists(dataset_name_or_path):
                    split_dir = os.path.join(dataset_name_or_path, split_name)
                else:
                    split_dir = os.path.join(work_dir, dataset_name_or_path, split_name)
                split_files = glob.glob(os.path.join(split_dir, '**', '*.json'))
                split_data = []
                for file_path in split_files:
                    if os.path.exists(file_path):
                        with open(file_path, 'r') as f:
                            split_data.append(json.load(f))
                if subset_name in data_dict:
                    data_dict[subset_name].update({split_name: split_data})
                else:
                    data_dict[subset_name] = {split_name: split_data}

        return data_dict

    def gen_prompt(self, input_d: dict, few_shot_list: list, **kwargs) -> dict:
        """
        Generate the prompt for the model input.

        Args:
            input_d: raw input dict.
                {"problem": "How many vertical asymptotes does the graph of $y=\\frac{2}{x^2+x-6}$ have?", "level": "Level 3", "type": "Algebra", "solution": "The denominator of the rational function factors into $x^2+x-6=(x-2)(x+3)$. Since the numerator is always nonzero, there is a vertical asymptote whenever the denominator is $0$, which occurs for $x = 2$ and $x = -3$.  Therefore, the graph has $\\boxed{2}$ vertical asymptotes."}

            few_shot_list:  few shot list. Each item is a raw input dict.
            **kwargs:

        Returns:
            {'data': [prompt]}
        """
        use_fewshot = self.few_shot_num > 0
        full_prompt = self._generate_prompt(input_d, use_fewshot=use_fewshot)

        return {'data': [full_prompt]}

    def get_gold_answer(self, input_d: dict) -> str:
        # Extract the gold answer from the input dict.
        return self._preprocess_input(input_d['solution'])

    def parse_pred_result(self, result: str, raw_input_d: dict = None, eval_type: str = 'checkpoint') -> str:
        """
        Parse the model output to get the answer. Could be the best choice index.

        Args:
            result: Predicted answer from the model. Usually a string for chat.
            raw_input_d (dict): The raw input. Depending on the dataset.
            eval_type: 'checkpoint' or 'service' or `custom`

        Returns:
            The parsed answer. Depending on the dataset. Usually a string for chat.
        """
        # TODO: check answer extraction
        # Note: Use same extraction method for both of checkpoint/service/custom
        return self._math_postprocess(result)

    def match(self, gold: str, pred: str) -> float:
        res = 0
        if self._is_equiv(pred, gold):
            res = 1

        return res

    def compute_metric(self, review_res_list: list) -> float:
        """
        Compute evaluation result by specific metric.

        Args:
            review_res_list: review score list, e.g. [0, 1, 1, 0, ...]

        Returns:
            The metric score.
        """
        items = [(score, 1.0) for score in review_res_list]
        return weighted_mean(items)

    def gen_report(self, subset_score_map: dict, report_name: str = None) -> dict:
        """
        Generate the report for the model output.

        Args:
            subset_score_map: The subset-score mapping. e.g. {subset_name: (score, num), ...}
            report_name: The user-defined report name.

        Returns: A dict of metric calculation results. The format is like:
        {
            "name":"CompetitionMath",
            "metric":"WeightedAverageAccuracy",
            "score":0.5632,
            "category":[
                {
                    "name":"DEFAULT",
                    "score":0.5632,
                    "subset":[
                        {
                            "name":"main",
                            "score":0.5632
                        },
                    ]
                }
            ],
            "total_num":100
        }
        """
        total_num: int = sum([num for _, num in subset_score_map.values()])
        weighted_avg_acc: float = sum([score * num for score, num in subset_score_map.values()]) / total_num
        weighted_avg_acc = normalize_score(score=weighted_avg_acc)
        cate_avg_list = [{'name': subset_name, 'score': normalize_score(score=score)} for subset_name, (score, _) in subset_score_map.items()]

        category_d = dict(name='DEFAULT',
                          score=weighted_avg_acc,
                          subset=cate_avg_list)

        res_map = dict(name=report_name or 'competition_math',
                       metric=self.metric_list[0]['name'],
                       score=weighted_avg_acc,
                       category=[category_d],
                       total_num=total_num)

        return res_map

    @classmethod
    def _generate_prompt(cls, input_d: dict, use_fewshot: bool = True) -> str:
        problem: str = input_d['problem']

        if use_fewshot:
            # Use 4-shot examples by system
            context = (
                'Problem:\nFind the domain of the expression $\\frac{{\sqrt{{x-2}}}}{{\sqrt{{5-x}}}}$.}}\nSolution:\nThe expressions inside each square root must be non-negative. Therefore, $x-2 \ge 0$, so $x\ge2$, and $5 - x \ge 0$, so $x \le 5$. Also, the denominator cannot be equal to zero, so $5-x>0$, which gives $x<5$. Therefore, the domain of the expression is $\\boxed{{[2,5)}}$.\nFinal Answer: The final answer is $[2,5)$. I hope it is correct.\n'
                'Problem:\nIf $\det \mathbf{{A}} = 2$ and $\det \mathbf{{B}} = 12,$ then find $\det (\mathbf{{A}} \mathbf{{B}}).$\nSolution:\nWe have that $\det (\mathbf{{A}} \mathbf{{B}}) = (\det \mathbf{{A}})(\det \mathbf{{B}}) = (2)(12) = \\boxed{{24}}.$\nFinal Answer: The final answer is $24$. I hope it is correct.\n'
                'Problem:\nTerrell usually lifts two 20-pound weights 12 times. If he uses two 15-pound weights instead, how many times must Terrell lift them in order to lift the same total weight?\nSolution:\nIf Terrell lifts two 20-pound weights 12 times, he lifts a total of $2\cdot 12\cdot20=480$ pounds of weight. If he lifts two 15-pound weights instead for $n$ times, he will lift a total of $2\cdot15\cdot n=30n$ pounds of weight. Equating this to 480 pounds, we can solve for $n$: \\begin{{align*}} 30n&=480\\\\ \Rightarrow\qquad n&=480/30=\\boxed{{16}} \end{{align*}}\nFinal Answer: The final answer is $16$. I hope it is correct.\n'
                'Problem:\nIf the system of equations: \\begin{{align*}} 6x-4y&=a,\\\\ 6y-9x &=b. \end{{align*}}has a solution $(x, y)$ where $x$ and $y$ are both nonzero, find $\\frac{{a}}{{b}},$ assuming $b$ is nonzero.\nSolution:\nIf we multiply the first equation by $-\\frac{{3}}{{2}}$, we obtain $$6y-9x=-\\frac{{3}}{{2}}a.$$Since we also know that $6y-9x=b$, we have $$-\\frac{{3}}{{2}}a=b\Rightarrow\\frac{{a}}{{b}}=\\boxed{{-\\frac{{2}}{{3}}}}.$$\nFinal Answer: The final answer is $-\\frac{{2}}{{3}}$. I hope it is correct.\n'
                f'Problem:\n{problem}\nSolution:\n'
            )
        else:
            context = 'Problem:\n' + problem + '\nSolution:\n'
        return context

    @classmethod
    def _preprocess_input(cls, input: str) -> str:
        """
        Preprocess the input data, remove the boxed solution.

        Args:
            input_d: The raw input. A single data format of the Competition Math.

        Returns:
            The preprocessed input.
        """
        return cls._remove_boxed(cls._last_boxed_only_string(input))

    @classmethod
    def _remove_boxed(cls, s):
        if s is None:
            return s

        if '\\boxed ' in s:
            left = '\\boxed '
            assert s[: len(left)] == left
            return s[len(left):]

        left = '\\boxed{'

        assert s[: len(left)] == left
        assert s[-1] == '}'

        return s[len(left): -1]

    @classmethod
    def _last_boxed_only_string(cls, string):

        idx = string.rfind('\\boxed')
        if '\\boxed ' in string:
            return '\\boxed ' + string.split('\\boxed ')[-1].split('$')[0]
        if idx < 0:
            idx = string.rfind('\\fbox')
            if idx < 0:
                return None

        i = idx
        right_brace_idx = None
        num_left_braces_open = 0
        while i < len(string):
            if string[i] == '{':
                num_left_braces_open += 1
            if string[i] == '}':
                num_left_braces_open -= 1
                if num_left_braces_open == 0:
                    right_brace_idx = i
                    break
            i += 1

        if right_brace_idx is None:
            retval = None
        else:
            retval = string[idx: right_brace_idx + 1]

        return retval

    @classmethod
    def _is_equiv(cls, str1, str2, verbose=False):
        if str1 is None and str2 is None:
            logger.warning('WARNING: Both None')
            return True
        if str1 is None or str2 is None:
            return False

        try:
            ss1 = cls.strip_string(str1)
            ss2 = cls.strip_string(str2)
            if verbose:
                logger.info(f'ss1: {ss1}, ss2: {ss2}')
            return ss1 == ss2
        except Exception:
            return str1 == str2

    @classmethod
    def strip_string(cls, string):
        # linebreaks
        string = string.replace('\n', '')

        # remove inverse spaces
        string = string.replace('\\!', '')

        # replace \\ with \
        string = string.replace('\\\\', '\\')

        # replace tfrac and dfrac with frac
        string = string.replace('tfrac', 'frac')
        string = string.replace('dfrac', 'frac')

        # remove \left and \right
        string = string.replace('\\left', '')
        string = string.replace('\\right', '')

        # Remove circ (degrees)
        string = string.replace('^{\\circ}', '')
        string = string.replace('^\\circ', '')

        # remove dollar signs
        string = string.replace('\\$', '')

        # remove units (on the right)
        string = cls.remove_right_units(string)

        # remove percentage
        string = string.replace('\\%', '')
        string = string.replace('\%', '')  # noqa: W605

        # " 0." equivalent to " ." and "{0." equivalent to "{." Alternatively, add "0" if "." is the start of the string
        string = string.replace(' .', ' 0.')
        string = string.replace('{.', '{0.')
        # if empty, return empty string
        if len(string) == 0:
            return string
        if string[0] == '.':
            string = '0' + string

        # to consider: get rid of e.g. "k = " or "q = " at beginning
        if len(string.split('=')) == 2:
            if len(string.split('=')[0]) <= 2:
                string = string.split('=')[1]

        # fix sqrt3 --> sqrt{3}
        string = cls.fix_sqrt(string)

        # remove spaces
        string = string.replace(' ', '')

        # \frac1b or \frac12 --> \frac{1}{b} and \frac{1}{2}, etc. Even works with \frac1{72} (but not \frac{72}1). Also does a/b --> \\frac{a}{b}
        string = cls.fix_fracs(string)

        # manually change 0.5 --> \frac{1}{2}
        if string == '0.5':
            string = '\\frac{1}{2}'

        # NOTE: X/Y changed to \frac{X}{Y} in dataset, but in simple cases fix in case the model output is X/Y
        string = cls.fix_a_slash_b(string)

        return string

    @classmethod
    def remove_right_units(cls, string):
        # "\\text{ " only ever occurs (at least in the val set) when describing units
        if '\\text{ ' in string:
            splits = string.split('\\text{ ')
            assert len(splits) == 2
            return splits[0]
        else:
            return string

    @classmethod
    def fix_fracs(cls, string):
        substrs = string.split('\\frac')
        new_str = substrs[0]
        if len(substrs) > 1:
            substrs = substrs[1:]
            for substr in substrs:
                new_str += '\\frac'
                if substr[0] == '{':
                    new_str += substr
                else:
                    try:
                        assert len(substr) >= 2
                    except AssertionError:
                        return string
                    a = substr[0]
                    b = substr[1]
                    if b != '{':
                        if len(substr) > 2:
                            post_substr = substr[2:]
                            new_str += '{' + a + '}{' + b + '}' + post_substr
                        else:
                            new_str += '{' + a + '}{' + b + '}'
                    else:
                        if len(substr) > 2:
                            post_substr = substr[2:]
                            new_str += '{' + a + '}' + b + post_substr
                        else:
                            new_str += '{' + a + '}' + b
        string = new_str
        return string

    @classmethod
    def fix_sqrt(cls, string):
        if '\\sqrt' not in string:
            return string
        splits = string.split('\\sqrt')
        new_string = splits[0]
        for split in splits[1:]:
            if split[0] != '{':
                a = split[0]
                new_substr = '\\sqrt{' + a + '}' + split[1:]
            else:
                new_substr = '\\sqrt' + split
            new_string += new_substr
        return new_string

    @classmethod
    def fix_a_slash_b(cls, string):
        if len(string.split('/')) != 2:
            return string
        a = string.split('/')[0]
        b = string.split('/')[1]
        try:
            a = int(a)
            b = int(b)
            assert string == '{}/{}'.format(a, b)
            new_string = '\\frac{' + str(a) + '}{' + str(b) + '}'
            return new_string
        except AssertionError:
            return string

    @classmethod
    def _math_postprocess(cls, text: str) -> str:
        SUBSTITUTIONS = [('an ', ''), ('a ', ''), ('.$', '$'), ('\\$', ''),
                         (r'\ ', ''), (' ', ''), ('mbox', 'text'),
                         (',\\text{and}', ','), ('\\text{and}', ','),
                         ('\\text{m}', '\\text{}'), ('\\le', '<')]
        REMOVED_EXPRESSIONS = [
            'square', 'ways', 'integers', 'dollars', 'mph', 'inches', 'ft',
            'hours', 'km', 'units', '\\ldots', 'sue', 'points', 'feet', 'minutes',
            'digits', 'cents', 'degrees', 'cm', 'gm', 'pounds', 'meters', 'meals',
            'edges', 'students', 'childrentickets', 'multiples', '\\text{s}',
            '\\text{.}', '\\text{\ns}', '\\text{}^2', '\\text{}^3', '\\text{\n}',
            '\\text{}', r'\mathrm{th}', r'^\circ', r'^{\circ}', r'\;', r',\!',
            '{,}', '"', '\\dots', '\n', '\r', '\f'
        ]
        import re

        def normalize_final_answer(final_answer: str) -> str:
            """Normalize a final answer to a quantitative reasoning question."""
            # final_answer = final_answer.split('=')[-1]
            for before, after in SUBSTITUTIONS:
                final_answer = final_answer.replace(before, after)
            for expr in REMOVED_EXPRESSIONS:
                final_answer = final_answer.replace(expr, '')

            # Extract answer that is in LaTeX math, is bold,
            # is surrounded by a box, etc.
            final_answer = re.sub(r'(\\text\{)(.*?)(\})', '\\2', final_answer)
            final_answer = re.sub(r'(\\textbf\{)(.*?)(\})', '\\2', final_answer)
            final_answer = re.sub(r'(\\overline\{)(.*?)(\})', '\\2', final_answer)
            final_answer = re.sub(r'(\\boxed\{)(.*)(\})', '\\2', final_answer)
            assert '\n' not in final_answer
            assert '\r' not in final_answer
            assert '\f' not in final_answer
            if len(re.findall(r'finalansweris(.*)', final_answer)) > 0:
                final_answer = re.findall(r'finalansweris(.*)', final_answer)[-1]

            if len(re.findall(r'oxed\{(.*?)\}', final_answer)) > 0:
                final_answer = re.findall(r'oxed\{(.*?)\}', final_answer)[-1]

            if len(re.findall(r'\$(.*?)\$', final_answer)) > 0:
                final_answer = re.findall(r'\$(.*?)\$', final_answer)[-1]
            final_answer = final_answer.strip()
            if 'rac' in final_answer and '\\frac' not in final_answer:
                final_answer = final_answer.replace('rac', '\\frac')

            final_answer = re.sub(r'(frac)([^{])(.)', 'frac{\\2}{\\3}',
                                  final_answer)
            final_answer = re.sub(r'(sqrt)([^{])', 'sqrt{\\2}', final_answer)
            final_answer = final_answer.replace('$', '')

            # Normalize 100,000 -> 100000
            if final_answer.replace(',', '').isdigit():
                final_answer = final_answer.replace(',', '')

            return final_answer

        for maybe_ans in text.split('.'):
            if 'final answer' in maybe_ans.lower():
                return normalize_final_answer(maybe_ans)
        return normalize_final_answer(text.split('.')[0])



import logging
import sys
from functools import lru_cache
from typing import Sequence

import json
import regex as re

logger = logging.getLogger(__name__)


def get_pairs(word):
    """Return set of symbol pairs in a word.

    Word is represented as tuple of symbols (symbols being variable-length strings).
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


@lru_cache()
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = list(range(ord('!'),
                    ord('~') + 1)) + list(range(
                        ord('¡'),
                        ord('¬') + 1)) + list(range(ord('®'),
                                                    ord('ÿ') + 1))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


class GPT2Tokenizer(object):
    """
    GPT-2 BPE tokenizer. Peculiarities:
        - Byte-level BPE

    [NOTE]: Copied from megatron.tokenizer.gpt2_tokenization.GPT2Tokenizer.
    """

    def __init__(self,
                 vocab_file,
                 merges_file,
                 errors='replace',
                 special_tokens=None,
                 max_len=None):
        assert sys.version_info[0] != 2

        self.max_len = max_len if max_len is not None else int(1e12)
        self.encoder = json.load(open(vocab_file))
        self.decoder = {v: k for k, v in self.encoder.items()}
        self.errors = errors  # how to handle errors in decoding
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        bpe_data = open(merges_file, encoding='utf-8').read().split('\n')[1:-1]
        bpe_merges = [tuple(merge.split()) for merge in bpe_data]
        self.bpe_ranks = dict(zip(bpe_merges, range(len(bpe_merges))))
        self.cache = {}

        # Should haved added re.IGNORECASE so BPE merges can happen for
        # capitalized versions of contractions
        self.pat = re.compile(
            r"""'s|'t|'re|'ve|'m|'ll|'d| ?\p{L}+| ?\p{N}+| ?[^\s\p{L}\p{N}]+|\s+(?!\S)|\s+"""
        )

        self.special_tokens = {}
        self.special_tokens_decoder = {}
        self.set_special_tokens(special_tokens)

    def __len__(self):
        return len(self.encoder) + len(self.special_tokens)

    def set_special_tokens(self, special_tokens):
        """ Add a list of additional tokens to the encoder.
            The additional tokens are indexed starting from the last index of the
            current vocabulary in the order of the `special_tokens` list.
        """
        if not special_tokens:
            self.special_tokens = {}
            self.special_tokens_decoder = {}
            return
        self.special_tokens = dict((tok, len(self.encoder) + i)
                                   for i, tok in enumerate(special_tokens))
        self.special_tokens_decoder = {
            v: k
            for k, v in self.special_tokens.items()
        }
        logger.info('Special tokens {}'.format(self.special_tokens))

    def bpe(self, token):
        if token in self.cache:
            return self.cache[token]
        word = tuple(token)
        pairs = get_pairs(word)

        if not pairs:
            return token

        while True:
            bigram = min(
                pairs, key=lambda pair: self.bpe_ranks.get(pair, float('inf')))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except BaseException:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[
                        i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = ' '.join(word)
        self.cache[token] = word
        return word

    def tokenize(self, text):
        """ Tokenize a string. """
        bpe_tokens = []
        for token in re.findall(self.pat, text):
            if sys.version_info[0] == 2:
                token = ''.join(self.byte_encoder[ord(b)] for b in token)
            else:
                token = ''.join(self.byte_encoder[b]
                                for b in token.encode('utf-8'))
            bpe_tokens.extend(
                bpe_token for bpe_token in self.bpe(token).split(' '))
        return bpe_tokens

    def convert_tokens_to_ids(self, tokens):
        """ Converts a sequence of tokens into ids using the vocab. """
        ids = []
        if isinstance(tokens, str):
            if tokens in self.special_tokens:
                return self.special_tokens[tokens]
            else:
                return self.encoder.get(tokens, 0)
        for token in tokens:
            if token in self.special_tokens:
                ids.append(self.special_tokens[token])
            else:
                ids.append(self.encoder.get(token, 0))
        if len(ids) > self.max_len:
            logger.warning(
                'Token indices sequence length is longer than the specified maximum '
                ' sequence length for this OpenAI GPT model ({} > {}). Running this'
                ' sequence through the model will result in indexing errors'.
                format(len(ids), self.max_len))
        return ids

    def convert_ids_to_tokens(self, ids, skip_special_tokens=False):
        """Converts a sequence of ids in BPE tokens using the vocab."""
        tokens = []
        for i in ids:
            if i in self.special_tokens_decoder:
                if not skip_special_tokens:
                    tokens.append(self.special_tokens_decoder[i])
            else:
                tokens.append(self.decoder[i])
        return tokens

    def encode(self, text):
        return self.convert_tokens_to_ids(self.tokenize(text))

    def decode(self, tokens):
        text = ''.join([self.decoder[token] for token in tokens])
        text = bytearray([self.byte_decoder[c] for c in text]).decode(
            'utf-8', errors=self.errors)
        return text


class DummyTokenizer:

    def tokenize(self, text: str):
        return text.split()


def get_tokenized_string(tokenizer: GPT2Tokenizer, text_list: Sequence[str]):
    token_ids_list, tokenized_string_list = [], []
    for text in text_list:
        assert tokenizer is not None
        token_ids = tokenizer.encode(text)
        tokenized_string = ' '.join(tokenizer.convert_ids_to_tokens(token_ids))
        token_ids_list.append(token_ids)
        tokenized_string_list.append(tokenized_string)
    return token_ids_list, tokenized_string_list



# Copyright (c) Alibaba, Inc. and its affiliates.
# flake8: noqa

import os
import random
import sys
import time
from abc import ABC, abstractmethod
from functools import partial
from typing import Any, List

import pandas as pd

from llmuses.constants import ArenaMode, EvalConfigKeys, FnCompletionParser, PositionBiasMitigation
from llmuses.models.openai_model import OpenAIModel
from llmuses.utils import completion_parsers
from llmuses.utils.arena_utils import (get_battle_pairs,
                                       merge_ques_ans,
                                       shuffle_pairwise_preferences)
from llmuses.utils import dump_jsonl_data, jsonl_to_list, random_seeded_choice
from llmuses.utils.logger import get_logger

logger = get_logger()


class BaseReviewer(ABC):

    def __init__(self, **kwargs):
        ...

    @abstractmethod
    def run(self, *args, **kwargs):
        """
        Run pairwise battles with given models.
        """
        raise NotImplementedError(
            'run() method must be implemented in your subclass.')


class AutoReviewerGpt4(BaseReviewer):
    """
    Auto-review target answers(models) pairwise with GPT-4.

    Args:
        prompt_file: path to prompt templates file.
        answer_file_list: list of paths to answer files.
        review_result_file: path to review result file.
        reviewer_args: config for reviewer(GPT-4).

    Examples:
        >>> from llmuses.evaluator.reviewer.auto_reviewer import AutoReviewerGpt4
        >>> input_kwargs = dict(prompt_file='/path/to/prompt_file.jsonl', answer_file_list=['/path/to/ans1_file.jsonl',
            '/path/to/ans2_file.jsonl', ...], review_file='/path/to/review_file.jsonl',
            reviewer_args={'model': 'gpt-4', 'mode': 'single'})
        >>> auto_reviewer = AutoReviewerGpt4(**input_kwargs)
        >>> auto_reviewer.run(dry_run=False)
    """

    MODEL_NAME = 'gpt-4'

    def __init__(self,
                 prompt_file: str,
                 answer_file_list: list,
                 review_result_file: str,
                 baseline_file: str = None,
                 reference_file: str = None,
                 reviewer_args: dict = None,
                 cache_file: str = None,
                 **kwargs):
        super().__init__(**kwargs)

        self.review_result_file = review_result_file
        self.prompt_list = jsonl_to_list(prompt_file)
        self.answer_list = [
            jsonl_to_list(answer_file) for answer_file in answer_file_list
        ]
        self.reference_list = jsonl_to_list(
            reference_file) if reference_file else []
        self.cache_list = jsonl_to_list(
            cache_file) if cache_file and os.path.isfile(cache_file) else []

        self.reviewer_args = reviewer_args if reviewer_args \
            else self._get_default_args()

        self.review_mode = self.reviewer_args.pop('mode', ArenaMode.PAIRWISE)
        if self.review_mode == ArenaMode.PAIRWISE_BASELINE:
            assert baseline_file is not None, f'baseline_file is required for {ArenaMode.PAIRWISE_BASELINE} mode'
            self.answer_list.append(jsonl_to_list(baseline_file))
            self.baseline_idx = len(self.answer_list) - 1

        self.position_bias_mitigation = self.reviewer_args.pop(
            EvalConfigKeys.POSITION_BIAS_MITIGATION,
            PositionBiasMitigation.NONE)
        if self.position_bias_mitigation == PositionBiasMitigation.RANDOMIZE_ORDER:
            self.random_seed = self.reviewer_args.pop(
                EvalConfigKeys.RANDOM_SEED, 123)

        fn_completion_parser = self.reviewer_args.pop(
            EvalConfigKeys.FN_COMPLETION_PARSER,
            FnCompletionParser.LMSYS_PARSER)
        completion_parser_kwargs = self.reviewer_args.pop(
            EvalConfigKeys.COMPLETION_PARSER_KWARGS, {})
        if isinstance(fn_completion_parser, str):
            fn_completion_parser = getattr(completion_parsers,
                                           fn_completion_parser)

        self.fn_completion_parser = partial(fn_completion_parser,
                                            **completion_parser_kwargs)
        self.gpt_predictor = OpenAIModel(model_cfg=self.reviewer_args)

    @staticmethod
    def _get_default_args():
        return dict(
            model=AutoReviewerGpt4.MODEL_NAME,
            mode=ArenaMode.PAIRWISE,
            position_bias_mitigation=PositionBiasMitigation.NONE,
            fn_completion_parser=FnCompletionParser.LMSYS_PARSER,
            random_seed=123,
        )

    @staticmethod
    def gen_prompt(prompts_list: list,
                   type: str,
                   category: str,
                   ques: str,
                   ans1: str,
                   ans2: str = None,
                   ans_ref: str = None):
        """
        Generate prompt for Auto-reviewer with GPT-4.
        """

        # Default to general category (idx 0)
        target_prompt_dict = prompts_list[0]
        for item in prompts_list:
            is_category_match = category in item['category'] if isinstance(
                item['category'], list) else item['category'] == category
            is_type_match = item.get('type', ArenaMode.PAIRWISE) == type
            if is_category_match and is_type_match:
                target_prompt_dict = item
                break
            elif is_type_match and target_prompt_dict.get('type',
                                                          ArenaMode.PAIRWISE) != type:
                target_prompt_dict = item  # fallback to type match

        sys_prompt = target_prompt_dict['system_prompt']
        prompt_template = target_prompt_dict['prompt_template']
        defaults = target_prompt_dict.get('defaults', dict({}))
        output_format = target_prompt_dict.get('output_format',
                                               '[[rating_a,rating_b]]')

        if type == ArenaMode.SINGLE:
            user_prompt = prompt_template.format(
                question=ques, answer=ans1, ref_answer_1=ans_ref, **defaults)
        else:
            user_prompt = prompt_template.format(
                question=ques,
                answer_a=ans1,
                answer_b=ans2,
                ref_answer_1=ans_ref,
                **defaults)

        return sys_prompt, user_prompt, output_format

    def get_review_cache(self, model_a, model_b, question) -> list:
        if model_b:
            cache_hit = next(
                (r for r in self.cache_list if r['model_a'] == model_a
                 and r['model_b'] == model_b and r['question'] == question),
                None)
        else:
            cache_hit = next(
                (r for r in self.cache_list
                 if r['model'] == model_a and r['question'] == question), None)
        return cache_hit

    def get_review_pair(self, item: List[dict], dry_run=False, **kwargs) -> dict:

        question = item[0]['text']
        question_id = item[0]['question_id']
        category = item[0]['category']

        model_a = item[0]['model_id']
        model_b = item[1]['model_id']

        ans1 = item[0]['answer']
        ans2 = item[1]['answer']

        review_cache = self.get_review_cache(model_a, model_b, question)
        if review_cache:
            logger.info(f'Use cache review for {model_a} vs {model_b} ...')
            return review_cache

        if self.position_bias_mitigation == PositionBiasMitigation.SWAP_POSITION:
            review_text_1, winner_1, score_1 = self._get_review_pair(
                model_a, model_b, question, category, ans1, ans2, dry_run=dry_run, **kwargs)
            review_text_2, winner_2, score_2 = self._get_review_pair(
                model_b, model_a, question, category, ans2, ans1, dry_run=dry_run, **kwargs)

            # Swap winner for the second round.
            if winner_2 == 'model_a':
                winner_2 = 'model_b'
            elif winner_2 == 'model_b':
                winner_2 = 'model_a'
            review_result = dict(
                model_a=model_a,
                model_b=model_b,
                win_1=winner_1,
                win_2=winner_2,
                anony=True,
                tstamp=time.time(),
                language=item[0].get('language', 'NA'),
                question_id=question_id,
                category=category,
                question=question,
                review_text_1=review_text_1,
                review_text_2=review_text_2)
        else:
            review_text, winner, scores = self._get_review_pair(
                model_a, model_b, question, category, ans1, ans2, dry_run=dry_run, **kwargs)

            if dry_run:
                scores = [round(random.random(), 1), round(random.random(), 1)]
                winner = 'model_a' if scores[0] > scores[1] else 'model_b'

            review_result = dict(
                model_a=model_a,
                model_b=model_b,
                scores=scores,
                win=winner,
                anony=True,
                tstamp=time.time(),
                language=item[0].get('language', 'NA'),
                question_id=question_id,
                category=category,
                question=question,
                review_text=review_text)
        return review_result

    def get_review_single(self, row: List[dict], dry_run: bool = False, **kwargs):
        item = row[0]
        model = item['model_id']
        question = item['text']
        question_id = item['question_id']
        category = item['category']
        answer = item['answer']

        review_cache = self.get_review_cache(model, None, question)
        if review_cache:
            logger.info(f'Use cache review for {model} ...')
            return review_cache

        review_text, score = self._get_review_single(model, question, category, answer, dry_run=dry_run, **kwargs)

        review_result = dict(
            model=model,
            score=score,
            anony=True,
            tstamp=time.time(),
            language=item.get('language', 'NA'),
            question_id=question_id,
            category=category,
            question=question,
            review_text=review_text)
        return review_result

    def _get_review_pair(self, model_a, model_b, question, category, ans1, ans2, dry_run=False, **kwargs) -> (str, Any):
        input_msg = dict(
            ques=question, category=category, ans1=ans1, ans2=ans2)

        if self.reference_list:
            ans_ref = next((ref for ref in self.reference_list
                            if ref.get('text') == question), None)
            assert ans_ref['answer']
            input_msg['ans_ref'] = ans_ref['answer']

        sys_prompt, user_prompt, output_format = AutoReviewerGpt4.gen_prompt(
            prompts_list=self.prompt_list,
            type=ArenaMode.SINGLE if self.review_mode == ArenaMode.SINGLE else ArenaMode.PAIRWISE,
            **input_msg)

        if dry_run:
            review_text = self._get_reviewer_prediction_dummy(sys_prompt, user_prompt, output_format)
        else:
            review_text = self._get_reviewer_prediction(sys_prompt, user_prompt, **kwargs)

        result = self.fn_completion_parser(
            review_text, output_format=output_format)
        if not isinstance(result, tuple):
            result = (result, None)
        return review_text, *result

    def _get_review_single(self, model, question, category, answer, dry_run=False, **kwargs) -> (str, Any):
        input_msg = dict(ques=question, category=category, ans1=answer)

        if self.reference_list:
            ans_ref = next((ref for ref in self.reference_list
                            if ref.get('text') == question), None)
            assert ans_ref['answer']
            input_msg['ans_ref'] = ans_ref['answer']

        sys_prompt, user_prompt, output_format = AutoReviewerGpt4.gen_prompt(
            prompts_list=self.prompt_list,
            type=ArenaMode.SINGLE if self.review_mode == ArenaMode.SINGLE else ArenaMode.PAIRWISE,
            **input_msg)

        if dry_run:
            review_text = self._get_reviewer_prediction_dummy(sys_prompt, user_prompt, output_format)
        else:
            review_text = self._get_reviewer_prediction(sys_prompt, user_prompt, **kwargs)

        score = self.fn_completion_parser(review_text, output_format)
        return review_text, score

    def _get_reviewer_prediction_dummy(self, sys_prompt: str, user_prompt: str,
                                       output_format) -> str:
        logger.info('Get dummy scores for input prompt ...')
        if output_format == '[[rating]]':
            return f'[[{round(random.random(), 2)}]]'
        if output_format == '[[rating_a,rating_b]]':
            ans_list = [round(random.random(), 2), round(random.random(), 2)]
            return ' '.join(str(element) for element in ans_list)
        elif output_format == '[[A]]':
            return random.choice(['[[A]]', '[[B]]', '[[C]]'])
        elif output_format == "[{'model': <model-name>, 'rank': <model-rank>}, " \
                              "{'model': <model-name>, 'rank': <model-rank>}]":
            rank_1 = random.choice([1, 2])
            rank_2 = 1 if rank_1 == 2 else 2
            return f"[{{'model': 'model_a', 'rank': {rank_1}}}, {{'model': 'model_b', 'rank': {rank_2}}}]"

    def _get_reviewer_prediction(self, sys_prompt: str, user_prompt: str, **kwargs) -> str:

        input_msg = dict(sys_prompt=sys_prompt, user_prompt=user_prompt)

        # Call GPT-4 predictor
        # TODO: Add more reviewer implementation
        resp = self.gpt_predictor.predict(model_id=self.MODEL_NAME, inputs=input_msg, **kwargs)

        if resp is None or len(resp) == 0:
            logger.error(f'Failed to get response from {self.MODEL_NAME} for input: {input_msg}')

        ans_text = resp['ans_text']
        # model_id = resp['model_id']

        return ans_text

    def run(self, dry_run: bool = False, **kwargs):
        print(f'Run battles for models with dry_run={dry_run} ...')

        os.makedirs(os.path.dirname(self.review_result_file), exist_ok=True)

        if len(self.answer_list) == 0:
            raise Exception('The answer list cannot be empty.')

        merge_key = 'question_id'
        merged_ans_df = merge_ques_ans(self.answer_list, merge_key=merge_key)
        merged_ans_df = merged_ans_df.drop(columns=['question_id'])

        if self.review_mode == ArenaMode.PAIRWISE:
            battle_pairs = get_battle_pairs(merged_ans_df.columns)
        elif self.review_mode == ArenaMode.PAIRWISE_BASELINE:
            battle_pairs = get_battle_pairs(merged_ans_df.columns,
                                            self.baseline_idx)
        elif self.review_mode == ArenaMode.SINGLE:
            battle_pairs = [(col, ) for col in merged_ans_df.columns]
        else:
            raise Exception(f'NotSupported mode: {self.review_mode}')

        res_list = []
        for t in battle_pairs:
            pair_df = merged_ans_df[list(t)]
            if self.position_bias_mitigation == PositionBiasMitigation.RANDOMIZE_ORDER:
                pair_df.columns = ['output_1', 'output_2']
                pair_df['is_switched_outputs'] = pair_df.apply(
                    lambda x: random_seeded_choice(
                        seed='is_switched_outputs' + x[0]['text'] + str(
                            self.random_seed),
                        choices=[False, True],
                    ),
                    axis=1,
                )
                pair_df = shuffle_pairwise_preferences(
                    pair_df, pair_df['is_switched_outputs'])

            for index, row in pair_df.iterrows():
                row_result = self.get_review_pair(row.to_list(), dry_run=dry_run, **kwargs) \
                    if self.review_mode != ArenaMode.SINGLE \
                    else self.get_review_single(row.to_list(), dry_run=dry_run, **kwargs)
                res_list.append(row_result)
        dump_jsonl_data(res_list, self.review_result_file)


if __name__ == '__main__':
    from pathlib import Path

    work_path = os.path.join(Path(__file__).absolute().parent, '../../../')
    prompt_template_path = os.path.join(work_path, 'llmuses/registry/data/prompt_template/prompt_templates.jsonl')
    answer_file_list = [os.path.join(work_path, 'outputs/arena/default/answers/answer_chatglm2-6b.jsonl'),
                        os.path.join(work_path, 'outputs/arena/default/answers/answer_llama2-7b.jsonl')]
    review_result_file_path = os.path.join(work_path, 'outputs/arena/default/reviews/review_gpt4.jsonl')

    input_kwargs = dict(prompt_file=prompt_template_path,
                        answer_file_list=answer_file_list,
                        review_result_file=review_result_file_path,
                        reviewer_args={},
                        baseline_file='',
                        reference_file='',
                        cache_file='', )

    auto_reviewer = AutoReviewerGpt4(**input_kwargs)
    auto_reviewer.run(dry_run=True)



"""LLM performance benchmark client.
"""
import argparse
import asyncio
import json
import sqlite3, os
import time
import base64
import pickle
import importlib.util
import sys
from typing import List, Dict
from datetime import datetime, timezone
import aiohttp
from http import HTTPStatus
from server_sent_event import ServerSentEvent
import aiohttp
import numpy as np

from _logging import logger

_current_path = os.path.dirname(os.path.abspath(__file__))

_query_send_completed = False
_data_process_completed = False
_table_name = "result"   

async def on_request_start(session, context, params):
    logger.info(f'Starting request: <{params}>')

async def on_request_chunk_sent(session, context, params):
    logger.info(f'Request body: {params}')
    
async def on_response_chunk_received(session, context, params):
    logger.info(f'Response info: <{params}>')

class AioHttpClient:
    def __init__(self, 
                 url: str, 
                 conn_timeout: int = 120, 
                 read_timeout: int = 120,
                 headers: Dict = None,
                 debug: bool = False):
        # one client only has one connection
        client_timeout = aiohttp.ClientTimeout(total=read_timeout + conn_timeout,
                                               connect=conn_timeout,
                                               sock_read=read_timeout)
        if debug:
            trace_config = aiohttp.TraceConfig()
            trace_config.on_request_start.append(on_request_start)
            trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
            # not support server sent event(stream=true)
            trace_config.on_response_chunk_received.append(on_response_chunk_received)
        self.client = aiohttp.ClientSession(trace_configs= [trace_config] if debug else [], 
                                            connector=aiohttp.TCPConnector(limit=1),
                                            timeout=client_timeout)
        ua = "modelscope_bench"
        self.headers = {"user-agent": ua}
        if headers:
            self.headers.update(headers)
        self.url = url

    async def __aenter__(self):
       pass

    async def __aexit__(self, exc_type, exc, tb):
        await self.client.close()

    async def aio_call(self):
        response = self._handle_request()
        if self.stream:
            return (item async for item in response)
        else:
            result = await response.__anext__()
            try:
                await response.__anext__()
            except StopAsyncIteration:
                pass
            return result

    async def _handle_stream(self, response):
        is_error = False
        status_code = response.status
        async for line in response.content:
            if line:
                line = line.decode("utf8")
                line = line.rstrip("\n").rstrip("\r")
                sse_msg = ServerSentEvent.decode(line)
                if not sse_msg:
                    continue
                if sse_msg.event and sse_msg.event == "error": # dashscope error
                    is_error = True

                if sse_msg.data:
                    if sse_msg.data.startswith("[DONE]"): # openai api completed
                        break
                    yield (is_error, status_code, sse_msg.data)
                    # yield data

    async def _handle_response(self, response: aiohttp.ClientResponse):
        if (response.status == HTTPStatus.OK and "text/event-stream" in response.content_type):
            async for is_error, status_code, data in self._handle_stream(response):
                yield (is_error, status_code, data)
        elif response.status == HTTPStatus.OK:
            content = await response.read()
            yield (False, HTTPStatus.OK, content)
        else:
            if "application/json" in response.content_type:
                error = await response.json()
                yield (True, response.status, error)
            else:
                msg = await response.read()
                yield (True, response.status, msg)

    async def post(self, body):
        try:
            headers = {"Content-Type": "application/json", **self.headers}
            response = await self.client.request("POST",
                                             url=self.url,
                                             json=body,
                                             headers=headers)
            async with response:
                async for rsp in self._handle_response(response):
                    yield rsp
        except aiohttp.ClientConnectorError as e:
            logger.error(e)
            raise e
        except Exception as e:
            logger.error(e)
            raise e

def dynamic_import_module(input_output_process_file_path: str):
    """Dynamic import input output process python file.

    Args:
        input_output_process_file_path (str): The absolute path of the 
            input output process python path, or name of the format, 
            system support openai, dashscope format.
    """
    module_name = 'input_output_process'

    input_output_spec = importlib.util.spec_from_file_location(module_name, input_output_process_file_path)
    input_output_process_module = importlib.util.module_from_spec(input_output_spec)
    sys.modules[module_name] = input_output_process_module
    input_output_spec.loader.exec_module(input_output_process_module)
    return input_output_process_module

def get_input_output_processor(input_output_format: str):
    if input_output_format in ['vllm_qwen_openai_completion', 'dashscope_message']:
        return dynamic_import_module(os.path.join(_current_path, '%s.py'%input_output_format))
    else:
        return dynamic_import_module(input_output_format)
        
async def dispatch_requests_worker(request_queue: asyncio.Queue, args):
    input_output_processor = get_input_output_processor(args.format)
    input_requests = []
    if args.prompt is not None:
        if args.prompt.startswith("@"): # read local as prompt, same as curl --data
            with open(args.prompt, 'r', encoding='utf-8') as f:
                prompt = f.read()
        else:
            prompt = args.prompt
        input_requests = [input_output_processor.get_query(model=args.model,
                                                prompt=prompt,
                                                **(args.parameters))]*args.number
    elif args.dataset is not None:
        input_prompts = []
        min_prompt_length = 0
        max_prompt_length = 1000000000
        if args.max_prompt_length is not None:
            max_prompt_length = args.max_prompt_length
        if args.min_prompt_length is not None:
            min_prompt_length = args.min_prompt_length
        if args.prompt_generator is not None:
            prompt_generator_module = dynamic_import_module(args.prompt_generator)
            for prompt in prompt_generator_module.prompt_generator(args.dataset):
                input_prompts.append(prompt)
        else: # default read file line by line
            with open(args.dataset, "r", encoding="UTF-8") as f:
                input_prompts = f.readlines()
        input_requests = []
        for prompt in input_prompts:
            prompt_length = len(prompt)
            if prompt_length > min_prompt_length and prompt_length < max_prompt_length:
                input_requests.append(input_output_processor.get_query(model=args.model,
                                                    prompt=prompt,
                                                    **(args.parameters)))
                if len(input_requests) >= args.number:
                    break
    if not input_requests:
        raise "Not input querys"

    total_query_counter = 0
    for request in input_requests:
        await request_queue.put(request)
        if args.rate == float("inf"):
            # If the request rate is infinity, then we don't need to wait.
            continue
        # Sample the request interval from the exponential distribution.
        interval = np.random.exponential(1.0 / args.rate)
        # The next request will be sent after the interval.
        await asyncio.sleep(interval)
        total_query_counter += 1
    return total_query_counter


class BenchmarkData(dict):
    """Benchmark info, two parts
    1. query info.
       prompt length
    2. response info
       start send time
       list of package_receive_time
           package info.
       response complete time
           total response info(response tokens)       
    """
    def __init__(self):
        pass 


def calculate_query_stream_metric(benchmark_data):
    first_chunk_latency = benchmark_data["chunk_times"][0] - benchmark_data["start_time"] # the first chunk latency
    n_chunks = len( benchmark_data["chunk_times"]) - 2 # minus first and last chunk.
    n_chunks_time = benchmark_data["chunk_times"][-2] - benchmark_data["chunk_times"][0] # -2 to last chunk
    return (first_chunk_latency, n_chunks, n_chunks_time)

async def statistic_benchmark_metric_worker(benchmark_data_queue: asyncio.Queue, args):
    """Statistics of performance metrics based on performance data
    """
    n_succeed_queries = 0
    n_failed_queries = 0
    total_first_chunk_latency = 0
    total_latency = 0.0
    n_total_chunks = 0
    n_total_prompt_tokens = 0
    n_total_completion_tokens = 0
    qps = 0
    concurrency = args.parallel
    start_time = None
    total_chunks_time = 0.0
    avg_latency = -1
    avg_first_chunk_latency = -1
    avg_token_per_seconds = -1
    avg_time_per_token = -1
    n_avg_chunks = -1
    avg_chunk_time = -1
    avg_prompt_tokens =-1
    avg_completion_tokens = -1
    total_time = 1 # avoid divide by zero
    n_total_queries = 0
    # avg generate tps generated tokens / time 
    # avg chunk time, first latency - avg_chunk_time == first latency, 去掉第一个和最后一个，第一个和prefill合并了，最后一个生成token可能比较短
    # avg prefill tps
    # prefill time = 首包时间-avg_chunk_time
    # n-tokens-per-trunk 
    n_benchmark_result = 0
    response_parser = get_input_output_processor(args.format)
    if args.result_file:
        result_db_path = args.result_file
    else:
        utc_dt = datetime.now(timezone.utc)
        current_time = utc_dt.astimezone().strftime("%Y_%m_%d_%H_%M_%S_%f")
        result_db_path = "./benchmark_%s.db"%current_time
    con = sqlite3.connect(result_db_path)
    
    db_cur = con.cursor()
    # create table      
    # TPS output tokens per second
    # tpot Time per ooutput token
    db_cur.execute("CREATE TABLE %s(request, start_time, chunk_times, success, \
                   response_messages, completed_time, latency, first_chunk_latency, \
                   n_chunks, chunk_time, prompt_tokens, completion_tokens)"%_table_name)
    if args.wandb_api_key is not None:
        import wandb
        name = args.wandb_name if args.wandb_name is not None else '%s_%s' % (args.model, current_time)
        wandb.init(
        project="perf_benchmark",   
        name=name,     
        # track run metadata
        config={
        "model": args.model,
        "time": current_time
        })
        os.environ["WANDB_SILENT"] = "true"
        
    while True:
        try:
            benchmark_data = benchmark_data_queue.get_nowait()                
            benchmark_data_queue.task_done()
            n_benchmark_result += 1
        except asyncio.QueueEmpty as e:
            if _data_process_completed:
                break
            await asyncio.sleep(1)
            continue
        if start_time is None:
            start_time = benchmark_data["start_time"] # start time with first request start time
        # total requests
        total_time = time.perf_counter() - start_time
        
        if benchmark_data["success"]:
            n_succeed_queries += 1
            n_query_trunks = len(benchmark_data["chunk_times"])
            query_latency = benchmark_data["completed_time"] - benchmark_data["start_time"]
            if n_query_trunks > 1:
                query_first_chunk_latency, query_n_chunks, query_n_chunks_time = calculate_query_stream_metric(benchmark_data)
            else:
                query_first_chunk_latency = query_latency     # not stream mode, query latency is equal total latency
                query_n_chunks = 1
                query_n_chunks_time = query_latency
            
            n_query_prompt_tokens, n_query_completion_tokens = response_parser.parse_responses(benchmark_data["response_messages"])
            n_total_prompt_tokens += n_query_prompt_tokens
            n_total_completion_tokens += n_query_completion_tokens
            
            total_first_chunk_latency += query_first_chunk_latency
            total_latency += query_latency
            n_total_chunks += query_n_chunks
            total_chunks_time += query_n_chunks_time

            # calc average
            avg_first_chunk_latency = total_first_chunk_latency / n_succeed_queries
            # average latency
            avg_latency = total_latency / n_succeed_queries
            # average generate chunks
            if n_query_trunks > 1:
                n_avg_chunks = n_total_chunks / n_succeed_queries + 2     # we remove the frist and last chunk.
            else:
                n_avg_chunks = n_total_chunks / n_succeed_queries
            avg_chunk_time = total_chunks_time / n_total_chunks            
            avg_prompt_tokens = n_total_prompt_tokens / n_succeed_queries
            avg_completion_tokens = n_total_completion_tokens / n_succeed_queries
            # avg generate tps generated tokens / time 
            avg_token_per_seconds = n_total_completion_tokens / total_time
            avg_time_per_token = total_time / n_total_completion_tokens
            # save the benchmark data to database.
            # save data to dist.
            insert_sql = "INSERT INTO %s VALUES('%s', %s, '%s', '%s', '%s', %s, %s, %s, %s, %s, %s, %s)" % (
                _table_name,
                base64.b64encode(pickle.dumps(benchmark_data["request"])).decode("ascii"), 
                benchmark_data["start_time"],
                json.dumps(benchmark_data["chunk_times"]),
                benchmark_data["success"],
                base64.b64encode(pickle.dumps(benchmark_data["response_messages"])).decode("ascii"),
                benchmark_data["completed_time"],
                query_latency,
                query_first_chunk_latency,
                query_n_chunks,
                query_n_chunks_time,
                n_query_prompt_tokens,
                n_query_completion_tokens
            )        
        else:
            n_failed_queries += 1
            # save the benchmark data to database.
            # save data to dist.
            insert_sql = "INSERT INTO %s(request, start_time, chunk_times, success, response_messages, completed_time)\
                VALUES('%s', %s, '%s', '%s', '%s', %s)" % (
                _table_name,
                base64.b64encode(pickle.dumps(benchmark_data["request"])).decode("ascii"), 
                benchmark_data["start_time"],
                json.dumps(benchmark_data["chunk_times"]),
                benchmark_data["success"],
                base64.b64encode(pickle.dumps(benchmark_data["response_messages"])).decode("ascii"),
                benchmark_data["completed_time"]
            )  
        n_total_queries = float(n_succeed_queries + n_failed_queries) # float for calc
        qps = n_total_queries / total_time
        db_cur.execute(insert_sql)
        con.commit()
        default_ndigits = 3
        message = {"Time": round(total_time, default_ndigits), 
                   "concurrency": concurrency,
                   "completed": int(n_total_queries), 
                   "succeed": n_succeed_queries, 
                   "failed": n_failed_queries, 
                   "qps": round(qps, default_ndigits), 
                   "latency": round(avg_latency, default_ndigits), 
                   "time to first token": round(avg_first_chunk_latency, default_ndigits), 
                   "throughput(output tokens per second)": round(avg_token_per_seconds, default_ndigits), 
                   "time per output token": round(avg_time_per_token, 5), 
                   "package per request": round(n_avg_chunks, default_ndigits), 
                   "time per package": round(avg_chunk_time, default_ndigits), 
                   "input tokens per request": round(avg_prompt_tokens, default_ndigits), 
                   "output tokens per request": round(avg_completion_tokens, default_ndigits)}
        if args.wandb_api_key is not None:
            wandb.log(message)
        if int(n_total_queries) % args.log_every_n_query == 0:
            msg = json.dumps(message)
            msg = msg[1:-1].replace('"', '')
            logger.info(msg)
    con.commit()
    con.close()
    return (total_time, n_total_queries, 
            n_succeed_queries, n_failed_queries, 
            qps, avg_latency, avg_first_chunk_latency, 
            n_avg_chunks, avg_chunk_time, 
            avg_prompt_tokens, avg_completion_tokens, 
            avg_token_per_seconds, avg_time_per_token, 
            result_db_path)

def summary_result(expected_number_of_queries,
                   total_time, 
                   n_total_queries, 
                   n_succeed_queries, 
                   n_failed_queries, 
                   qps, 
                   avg_latency, 
                   avg_first_chunk_latency, 
                   n_avg_chunks, 
                   avg_chunk_time, 
                   avg_prompt_tokens, 
                   avg_completion_tokens, 
                   avg_token_per_seconds, 
                   avg_time_per_token,
                   result_db_path, args):
    
    print("Benchmarking summary: ")
    print(" Time taken for tests: %.3f seconds" % total_time)
    print(" Expected number of requests: %s" % expected_number_of_queries)
    print(" Number of concurrency: %d" % args.parallel)
    print(" Total requests: %d" % n_total_queries)
    print(" Succeed requests: %d" % n_succeed_queries)
    print(" Failed requests: %d" % n_failed_queries)
    print(" Average QPS: %.3f" % qps)
    print(" Average latency: %.3f" % avg_latency)
    print(" Throughput(average output tokens per second): %.3f" % avg_token_per_seconds)
    print(" Average time to first token: %.3f" % avg_first_chunk_latency)
    print(" Average input tokens per request: %.3f" % avg_prompt_tokens)
    print(" Average output tokens per request: %.3f" % avg_completion_tokens)
    print(" Average time per output token: %.5f" % avg_time_per_token)
    print(" Average package per request: %.3f" % n_avg_chunks)
    print(" Average package latency: %.3f" % avg_chunk_time)
    
    con = sqlite3.connect(result_db_path)
    query_sql = "SELECT start_time, chunk_times, success, \
                   completed_time, latency, first_chunk_latency, \
                   n_chunks, chunk_time, prompt_tokens, completion_tokens \
                       FROM %s WHERE success='True' ORDER BY first_chunk_latency ASC"%_table_name
    
    percentiles = [50, 66, 75, 80, 90, 95, 98, 99]
    with con:
        rows = con.execute(query_sql).fetchall()
        if len(rows) > 0:
            print(" Percentile of time to first token: ")
            for percentile in percentiles:
                idx = (int)(n_total_queries*percentile/100)
                row = rows[idx]
                print("     p%s: %.4f"%(percentile, row[5] if row[5] is not None else float("inf")))
                #print(row)
            print(" Percentile of request latency: ")
            latency_index = 4
            rows.sort(key=lambda x: x[latency_index])
            for percentile in percentiles:
                idx = (int)(n_total_queries*percentile/100)
                row = rows[idx]
                print("     p%s: %.4f"%(percentile, row[latency_index] if row[latency_index] is not None else float("inf")))        
    con.close()

async def send_requests_worker(task_id, request_queue: asyncio.Queue, benchmark_data_queue: asyncio.Queue, args):    
    client = AioHttpClient(args.url,
                           conn_timeout=args.connect_timeout,
                           read_timeout=args.read_timeout, 
                           headers=args.headers,
                           debug=args.debug)
    async with client:
        while True:        
            # Get a request out of the queue.
            try:
                request = request_queue.get_nowait()
                request_queue.task_done()
            except asyncio.QueueEmpty as e:
                if _query_send_completed:
                    break
                await asyncio.sleep(0.01)
                continue    # keep polling querys
            benchmark_data = BenchmarkData()
            benchmark_data["request"] = request
            benchmark_data["start_time"] = time.perf_counter()
            benchmark_data["chunk_times"] = []
            benchmark_data["success"] = False
            collected_messages = []
            try:
                async for (is_error, state_code, response_data) in client.post(request):
                    if is_error or state_code != HTTPStatus.OK:
                        logger.error("Request: %s failed, state_code: %s, data: %s"%(request, state_code, response_data.decode('utf-8')))
                        break
                    else:
                        if response_data:
                            collected_messages.append(response_data)  # save the message
                            logger.debug(response_data)
                            benchmark_data["chunk_times"].append(time.perf_counter())
                
                benchmark_data["response_messages"] = collected_messages
                benchmark_data["completed_time"] = time.perf_counter()
                benchmark_data["success"] = not is_error
                await benchmark_data_queue.put(benchmark_data)
            except Exception as e:
                if response_data:
                    collected_messages.append(response_data)  # save the message
                benchmark_data["response_messages"] = collected_messages
                benchmark_data["completed_time"] = time.perf_counter()
                await benchmark_data_queue.put(benchmark_data)
                logger.error("Request query: %s exception, response: %s" % (request, response_data))
                logger.error(e)

async def benchmark(args) -> None:
    request_tasks: List[asyncio.Task] = []
    # Queues can be used to distribute workload between several concurrent tasks
    # Create a queue that we will use to store our "workload".
    request_queue = asyncio.Queue()
    benchmark_data_queue = asyncio.Queue()
    dispatch_task = asyncio.create_task(dispatch_requests_worker(request_queue, args))
    statistic_benchmark_metric_task = asyncio.create_task(statistic_benchmark_metric_worker(benchmark_data_queue, args))
    for idx, task in enumerate(range(args.parallel)):
        task = asyncio.create_task(send_requests_worker(idx, request_queue, benchmark_data_queue, args))
        request_tasks.append(task)
    
   
    expected_number_of_queries = await dispatch_task # wait for dispatch task complete
    await request_queue.join()
    global _query_send_completed   
    _query_send_completed = True
    await asyncio.gather(*request_tasks, return_exceptions=True)
    await benchmark_data_queue.join() # wait for all query is processed
    global _data_process_completed
    _data_process_completed = True
    (total_time, n_total_queries, 
     n_succeed_queries, n_failed_queries, 
     qps, avg_latency, 
     avg_first_chunk_latency, n_avg_chunks, 
     avg_chunk_time, avg_prompt_tokens, 
     avg_completion_tokens, avg_token_per_seconds, 
     avg_time_per_token, result_db_path) = await statistic_benchmark_metric_task

    summary_result(expected_number_of_queries, total_time, n_total_queries, n_succeed_queries, 
            n_failed_queries, qps, avg_latency, avg_first_chunk_latency, 
            n_avg_chunks, avg_chunk_time,
            avg_prompt_tokens, avg_completion_tokens, 
            avg_token_per_seconds, avg_time_per_token, 
            result_db_path, args)
    await asyncio.sleep(0.250)

# from: https://gist.github.com/vadimkantorov/37518ff88808af840884355c845049ea
class ParseKVAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for each in values:
            try:
                key, value = each.split("=")
                getattr(namespace, self.dest)[key] = value
            except ValueError as ex:
                message = "\nTraceback: {}".format(ex)
                message += "\nError on '{}' || It should be 'key=value'".format(
                    each)
                raise argparse.ArgumentError(self, str(message))

def main(args):
    asyncio.run(benchmark(args))
    
def add_argument(parser: argparse.ArgumentParser):
    parser.add_argument("--model", type=str, required=True, 
                        help="The test model name.")
    parser.add_argument("--url", type=str, default="localhost")
    parser.add_argument("--dataset", type=str, required=False,
                        help="Path to the dataset, with prompt line by line")
    parser.add_argument("--connect-timeout", type=int, default=120,
                        help="The network connection timeout")
    parser.add_argument("--read-timeout", type=int, default=120,
                        help="The network read timeout")
    parser.add_argument("--max-prompt-length", type=int, required=False,
                        help="Maximum input prompt length")
    parser.add_argument("--min-prompt-length", type=int, required=False,
                        help="Minimum input prompt length.")
    parser.add_argument("--prompt", type=str, required=False,
                         help="Specified the request prompt, all the query will use this prompt.")
    parser.add_argument("-n", "--number", type=int, default=1000,
                        help="How many requests to be made.")
    parser.add_argument("--parameters", nargs="+", 
                        dest="parameters",
                        default={},
                        action=ParseKVAction,
                        help="Extra parameters accepts by key1=value1 key2=value2. "
                             "The parameters will be use for each query." 
                             "You can use this parameter to specify sample parameters such as top_p, top_k ",
                        metavar="KEY1=VALUE1")
    parser.add_argument("--headers", nargs="+", dest="headers",
                        action=ParseKVAction,
                        help="Extra http headers accepts by key1=value1 key2=value2. "
                             "The headers will be use for each query." 
                             "You can use this parameter to specify http auchorization and other header.",
                        metavar="KEY1=VALUE1")
    parser.add_argument("--parallel", type=int, default=1,
                         help="Set number of concurrency request, default 1")
    parser.add_argument("--rate", type=int, default=None,
                         help="Number of requests per second. default None, if it set to -1,"
                             "then all the requests are sent at time 0. "
                             "Otherwise, we use Poisson process to synthesize "
                             "the request arrival times.  Mutual exclusion with parallel")
    parser.add_argument("--log-every-n-query", type=int, default=10, 
                        help="Logging every n query.")
    parser.add_argument("--n-tokens-per-trunk",
                        type=int,
                        default=16,
                        help="Number of tokens per stream output trunk")
    parser.add_argument("--format",
                        type=str,
                        default="vllm_qwen_openai_completion",
                        help="Specify the request/response format, current support,"
                             "[vllm_qwen_openai_completion, dashscope_message], "
                             "you can define your input output format via python file,"
                             "please ref the vllm_qwen_openai_completion.py")
    parser.add_argument("--prompt-generator", type=str, default=None,
                        help="Specify the prompt generator python file, default \
                            we will read data line by line and prompt is one line")
    parser.add_argument("--result-file", type=str,
                        help="The path of the result file, default file in current directory name with current_time")
    parser.add_argument("--wandb-api-key", type=str, default=None,
                        help="The wandb api key, if set the metric will be saved to wandb.")
    parser.add_argument("--wandb-name", type=str, 
                        help="The one db result name, default: {model_name}_{current_time}")
    parser.add_argument("--debug", action='store_true', default=False,
                        help='Debug request send.')
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Benchmark large language model server performance.")
    add_argument(parser)
    args = parser.parse_args()
    main(args)




# Copyright (c) Alibaba, Inc. and its affiliates.


from llmuses.benchmarks.bbh.bbh_adapter import DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.bbh.bbh_adapter import BBHAdapter as DataAdapterClass
from llmuses.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass    # noqa



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright 2020 The HuggingFace Datasets Authors and the current dataset script contributor.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# flake8: noqa

"""Grade School Math 8k dataset."""

import json
import textwrap

import datasets


_CITATION = """\
@misc{cobbe2021training,
      title={Training Verifiers to Solve Math Word Problems},
      author={Karl Cobbe and Vineet Kosaraju and Mohammad Bavarian and Jacob Hilton and Reiichiro Nakano and Christopher Hesse and John Schulman},
      year={2021},
      eprint={2110.14168},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
}
"""

_DESCRIPTION = """\
GSM8K (Grade School Math 8K) is a dataset of 8.5K high quality
linguistically diverse grade school math word problems. The
dataset was created to support the task of question answering
on basic mathematical problems that require multi-step reasoning.
"""

_HOMEPAGE = 'https://openai.com/blog/grade-school-math'
_MODELSCOPE_PAGE = 'https://modelscope.cn/datasets/modelscope/gsm8k/summary'

_LICENSE = 'MIT'

# _BASE_URL = "https://raw.githubusercontent.com/openai/grade-school-math/master/grade_school_math/data/"
TRAIN_URL = 'https://sail-moe.oss-cn-hangzhou.aliyuncs.com/open_data/gsm8k/train.jsonl'
TEST_URL = 'https://sail-moe.oss-cn-hangzhou.aliyuncs.com/open_data/gsm8k/test.jsonl'


class Gsm8kConfig(datasets.BuilderConfig):
    """BuilderConfig for GSM8K."""

    def __init__(self, urls, **kwargs):
        """BuilderConfig for GSM8K.
        Args:
        urls: *dict[string]*, the urls for each split of the GSM8k set.
        """
        super().__init__(version=datasets.Version('1.1.0'), **kwargs)
        self.urls = urls


class Gsm8k(datasets.GeneratorBasedBuilder):
    """Grade School Math 8k (GSM8K)"""

    BUILDER_CONFIGS = [
        Gsm8kConfig(
            name='main',
            description=textwrap.dedent(
                """
                It is segmented into 7.5K training problems and 1K test problems.
                These problems take between 2 and 8 steps to solve, and solutions
                primarily involve performing a sequence of elementary calculations
                using basic arithmetic operations (+ - / *) to reach the final
                answer. A bright middle school student should be able to solve
                every problem.
                """,
            ),
            urls={
                'train': TRAIN_URL,
                'test': TEST_URL,
            },
        ),
    ]

    def _info(self):
        features = datasets.Features(
            {
                'question': datasets.Value('string'),
                'answer': datasets.Value('string'),
            }
        )
        return datasets.DatasetInfo(
            description=_DESCRIPTION,
            features=features,
            homepage=_HOMEPAGE,
            license=_LICENSE,
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        data_dir = dl_manager.download_and_extract(self.config.urls)
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    'filepath': data_dir['train'],
                },
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                gen_kwargs={
                    'filepath': data_dir['test'],
                },
            ),
        ]

    def _generate_examples(self, filepath):
        with open(filepath, encoding='utf-8') as f:
            for key, row in enumerate(f):
                data = json.loads(row)
                yield key, {
                    'question': data['question'],
                    'answer': data['answer'],
                }



# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.mmlu.mmlu_adapter import DATASET_ID, SUBJECT_MAPPING, SUBSET_LIST, MMLUAdapter
from llmuses.benchmarks.mmlu.mmlu_adapter import MMLUAdapter as DataAdapterClass
from llmuses.models.model_adapter import MultiChoiceModelAdapter as ModelAdapterClass    # noqa



# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.hellaswag.hellaswag_adapter import HellaSwagAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.hellaswag.hellaswag_adapter import HellaSwagAdapter as DataAdapterClass
from llmuses.models.model_adapter import ContinuationLogitsModelAdapter as ModelAdapterClass    # noqa



# Copyright (c) Alibaba, Inc. and its affiliates.
from enum import Enum

DEFAULT_ROOT_CACHE_DIR = '~/.cache/llmuses'


class DumpMode:
    OVERWRITE = 'overwrite'
    APPEND = 'append'


class MetricsConstant:
    EPSILON = float(1e-6)
    INVALID_VALUE = -9999999
    ROUGE_KEYS = [
        'rouge-1-r',
        'rouge-1-p',
        'rouge-1-f',
        'rouge-2-r',
        'rouge-2-p',
        'rouge-2-f',
        'rouge-l-r',
        'rouge-l-p',
        'rouge-l-f',
    ]


class MetricMembers(Enum):

    # Math accuracy metric
    MATH_ACCURACY = 'math_accuracy'

    # Code pass@k metric
    CODE_PASS_K = 'code_pass_k'

    # Code rouge metric
    ROUGE = 'rouge'

    # ELO rating system for pairwise comparison
    ELO = 'elo'

    # Pairwise comparison win/lose and tie(optional)
    PAIRWISE = 'pairwise'

    # Rating score for single model
    SCORE = 'score'


class ArenaWinner:

    MODEL_A = 'model_a'

    MODEL_B = 'model_b'

    TIE = 'tie'

    TIE_BOTH_BAD = 'tie_both_bad'

    UNKNOWN = 'unknown'


class ArenaMode:
    SINGLE = 'single'
    PAIRWISE = 'pairwise'
    PAIRWISE_BASELINE = 'pairwise_baseline'


class OutputsStructure:

    LOGS_DIR = 'logs_dir'

    PREDICTIONS_DIR = 'predictions_dir'

    REVIEWS_DIR = 'reviews_dir'

    REPORTS_DIR = 'reports_dir'

    CONFIGS_DIR = 'configs_dir'


class AnswerKeys:

    ANSWER_ID = 'answer_id'

    RAW_INPUT = 'raw_input'

    ORIGIN_PROMPT = 'origin_prompt'

    MODEL_SPEC = 'model_spec'

    SUBSET_NAME = 'subset_name'

    CHOICES = 'choices'


class ReviewKeys:

    REVIEW_ID = 'review_id'

    REVIEWED = 'reviewed'

    REVIEWER_SPEC = 'reviewer_spec'

    REVIEW_TIME = 'review_time'

    MESSAGE = 'message'

    CONTENT = 'content'

    GOLD = 'gold'

    PRED = 'pred'

    RESULT = 'result'

    REVIEW = 'review'


class EvalConfigKeys:
    CLASS_REF = 'ref'
    CLASS_ARGS = 'args'
    ENABLE = 'enable'
    POSITION_BIAS_MITIGATION = 'position_bias_mitigation'
    RANDOM_SEED = 'random_seed'
    FN_COMPLETION_PARSER = 'fn_completion_parser'
    COMPLETION_PARSER_KWARGS = 'completion_parser_kwargs'
    OUTPUT_FILE = 'output_file'
    MODEL_ID_OR_PATH = 'model_id_or_path'
    MODEL_REVISION = 'revision'
    GENERATION_CONFIG = 'generation_config'
    PRECISION = 'precision'


class FnCompletionParser:
    LMSYS_PARSER: str = 'lmsys_parser'
    RANKING_PARSER: str = 'ranking_parser'


class PositionBiasMitigation:
    NONE = 'none'
    RANDOMIZE_ORDER = 'randomize_order'
    SWAP_POSITION = 'swap_position'


class EvalStage:
    # Enums: `all`, `infer`, `review`
    ALL = 'all'
    INFER = 'infer'
    REVIEW = 'review'



# Copyright (c) Alibaba, Inc. and its affiliates.

from llmuses.benchmarks.competition_math.competition_math_adapter import CompetitionMathAdapter, DATASET_ID, SUBSET_LIST
from llmuses.benchmarks.competition_math.competition_math_adapter import CompetitionMathAdapter as DataAdapterClass
from llmuses.models.model_adapter import ChatGenerationModelAdapter as ModelAdapterClass    # noqa



# Copyright (c) Alibaba, Inc. and its affiliates.
# Copyright (c) lmsys.org.

import random
from collections import OrderedDict, defaultdict
from typing import List, Sequence, Union

import numpy as np
import pandas as pd
import pyarrow as pa

from llmuses.utils.logger import get_logger

logger = get_logger()


def compute_elo(battles,
                col_model_a='model_a',
                col_model_b='model_b',
                col_win='win',
                tie_values=['tie', 'tie (bothbad)'],
                k=32,
                scale=400,
                base=10,
                init_rating=1000):
    rating = defaultdict(lambda: init_rating)

    for rd, model_a, model_b, win in battles[[
            col_model_a, col_model_b, col_win
    ]].itertuples():
        ra = rating[model_a]
        rb = rating[model_b]
        ea = 1 / (1 + base**((rb - ra) / scale))
        eb = 1 / (1 + base**((ra - rb) / scale))
        if win == col_model_a:
            sa = 1
        elif win == col_model_b:
            sa = 0
        elif win in tie_values:
            sa = 0.5
        else:
            raise Exception(f'unexpected vote {win}')
        rating[model_a] += k * (sa - ea)
        rating[model_b] += k * (1 - sa - eb)

    return rating


def merge_ques_ans(answer_list_all,
                   merge_key: str = 'question_id',
                   merge_mode: str = 'inner') -> pd.DataFrame:
    """
    Merge question and answer list to unifiled data.

    Args:
        answer_list_all: list of answer list,
            e.g. [ans1_list, ans2_list, ...], an ans_list is predicted answers
            of a specific model, must contain following columns: 'question_id',
            'text', 'category', 'model_id', 'answer'
        merge_key: key for dataframe merging
        merge_mode: mode for dataframe merging,
            e.g. 'inner', 'left', 'right', 'outer'

    Returns:
        pandas DataFrame: merged dataframe, e.g. columns are
            ['question_id', 'gpt-3.5-turbo', 'llama2-7b']
    """
    ans_df = pd.DataFrame()
    for ans_list in answer_list_all:
        ans_list = [{
            'question_id': item['question_id'],
            item['model_id']: item
        } for item in ans_list]
        if ans_df.empty:
            ans_df = pa.Table.from_pylist(ans_list).to_pandas()
        else:
            ans_df = pd.merge(
                ans_df,
                pa.Table.from_pylist(ans_list).to_pandas(),
                on=merge_key,
                how=merge_mode)

    return ans_df


def get_battle_pairs(columns: List[str], baseline_idx: int = -1) -> List[tuple]:
    """
    Get battle pair names from columns.

    Args:
        columns: list of column names.

    Returns:
        list of battle pairs.

    Example:
        >>> columns = ['A', 'B', 'C']
        >>> res = get_battle_pairs(columns)
        >>> print(res)
        >>> [('B', 'A'), ('C', 'A'), ('C', 'B')]

        >>> columns = ['A', 'B', 'C']
        >>> res = get_battle_pairs(columns, 2)
        >>> print(res)
        >>> [('A', 'C'), ('B', 'C')]
    """
    res_list = []

    cols_num = len(columns)
    if cols_num <= 0:
        return res_list

    if baseline_idx != -1:
        n_column = columns[baseline_idx]
        res_list = [(column, n_column) for column in columns
                    if column != n_column]
    else:
        mat = np.ones((cols_num, cols_num))
        mat_lower_tril = np.tril(mat, k=-1)
        x_ids, y_ids = np.where(mat_lower_tril == 1)
        res_list = [(columns[x_id], columns[y_id])
                    for x_id, y_id in zip(x_ids, y_ids)]

    return res_list


def get_battle_pairs_origin(columns: List[str],
                            compare_base: bool = False,
                            swap: bool = False):  # TODO: to refactor
    """
    Get battle pair names from columns.

    Args:
        columns: list of column names.

    Returns:
        list of battle pairs.

    Example:
        >>> columns = ['A', 'B', 'C']
        >>> res = get_battle_pairs(columns)
        >>> print(res)
        >>> [('B', 'A'), ('C', 'A'), ('C', 'B')]
    """
    res_list = []

    cols_num = len(columns)
    if cols_num <= 0:
        return res_list

    if not compare_base:
        mat = np.ones((cols_num, cols_num))
        mat_lower_tril = np.tril(mat, k=-1)
        x_ids, y_ids = np.where(mat_lower_tril == 1)
        res_list = [(columns[x_id], columns[y_id])
                    for x_id, y_id in zip(x_ids, y_ids)]
    else:
        for column in columns[1:]:
            res_list.append((columns[0], column))

    if swap:
        res_list.extend([(j, i) for i, j in res_list])
    return res_list


def shuffle_pairwise_preferences(
        df: pd.DataFrame, arr_is_shuffle: Sequence[int]) -> pd.DataFrame:
    """Shuffle the outputs of a pairwise preference dataframe.

    Examples
    --------
    >>> df = pd.DataFrame([dict(instruction='2+2', output_1='3', output_2='4', preference=2),
                           dict(instruction='2+3', output_1='5', output_2='4', preference=1)])
    >>> print(shuffle_pairwise_preferences(df, [True, False]))
        instruction output_1 output_2  preference
    0         2+2        4        3           1
    1         2+3        5        4           1
    """
    col_1 = df['output_1'].copy()
    col_2 = df['output_2'].copy()
    df['output_1'] = np.where(arr_is_shuffle, col_2, col_1)
    df['output_2'] = np.where(arr_is_shuffle, col_1, col_2)

    if 'preference' in df.columns:
        df['preference'] = np.where(arr_is_shuffle, 3 - df['preference'],
                                    df['preference'])

    return df


class BattlePairSelection:
    """
    Select battle pairs by specific strategy.

    Attributes:
        model_elo_map(dict): map of model_id--base_elo_score
    """

    DEFAULT_K = 5

    def __init__(self, model_elo_map: Union[dict, OrderedDict]):
        # Make sure model_elo_map to be ordered when compare_base is true.
        self.model_elo_map = model_elo_map

    def top_k(self,
              k: int = DEFAULT_K,
              compare_base: bool = False,
              swap: bool = False) -> list:
        if k <= 0:
            k = self.DEFAULT_K
        sorted_res = sorted(self.model_elo_map.items(), key=lambda x: x[1])[:k]
        sorted_res = list(dict(sorted_res).keys())
        return get_battle_pairs_origin(sorted_res, compare_base, swap)

    def random_k(self,
                 k: int = DEFAULT_K,
                 compare_base: bool = False,
                 swap: bool = False) -> list:
        if k <= 0:
            k = self.DEFAULT_K
        if k > len(self.model_elo_map):
            k = len(self.model_elo_map)
        candidate_list = list(self.model_elo_map.items())
        k = len(candidate_list) if k > len(candidate_list) else k
        res = dict(random.sample(candidate_list, k=k))
        res = list(res.keys())
        return get_battle_pairs_origin(res, compare_base, swap)

    def volatility_index(self,
                         frac: float = 0.2,
                         compare_base: bool = False,
                         swap: bool = False) -> list:
        res_list = []
        candidate_list = get_battle_pairs_origin(
            list(self.model_elo_map.keys()), compare_base, swap)
        for t in candidate_list:
            model_a = t[0]
            model_b = t[1]
            base_elo_a = self.model_elo_map.get(model_a)
            base_elo_b = self.model_elo_map.get(model_b)

            vol_frac = abs(base_elo_b - base_elo_a) / max(
                base_elo_a, base_elo_b)
            if vol_frac <= frac:
                res_list.append(t)

        return res_list



# Copyright (c) Alibaba, Inc. and its affiliates.

import random
import time
from llmuses.models import ChatBaseModel
from llmuses.utils.logger import get_logger

logger = get_logger()


class DummyChatModel(ChatBaseModel):

    MODEL_ID = 'dummy_chat_model_0801'
    REVISION = 'v1.0.0'

    def __init__(self, model_cfg: dict, **kwargs):
        model_cfg['model_id'] = self.MODEL_ID
        model_cfg['revision'] = self.REVISION
        super(DummyChatModel, self).__init__(model_cfg=model_cfg)

    def predict(self, inputs: dict, **kwargs) -> dict:

        debug: bool = False
        if debug:
            messages = inputs['messages']
            history = inputs['history']

            logger.info(f'** messages: {messages}')
            logger.info(f'** history: {history}')

        choice = random.choice(['A', 'B', 'C', 'D'])

        # Build response
        res = {
            'choices': [
                {
                    'index': 0,
                    'message': {
                        'content': choice,
                        'role': 'assistant'
                    }
                }
            ],
            'created': time.time(),
            'model': self.MODEL_ID + '-' + self.REVISION,
            'object': 'chat.completion',
            'usage': {}
        }

        return res




import time
import jwt
import argparse
def generate_token(apikey: str, exp_seconds: int):
    try:
        id, secret = apikey.split(".")
    except Exception as e:
        raise Exception("invalid apikey", e)
 
    payload = {
        "api_key": id,
        "exp": int(round(time.time() * 1000)) + exp_seconds * 1000,
        "timestamp": int(round(time.time() * 1000)),
    }
 
    return jwt.encode(
        payload,
        secret,
        algorithm="HS256",
        headers={"alg": "HS256", "sign_type": "SIGN"},
    )
    
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate zhipu glm api token, you need install PyJWT with: pip install PyJWT")
    parser.add_argument("--api-key", type=str, required=True, 
                        help="The zhipu api key.")
    parser.add_argument("--exp-seconds", type=int, default=24*60*60,
                        help='The token expiration seconds')
    args = parser.parse_args()
    print('Token generated: %s' % generate_token(args.api_key, args.exp_seconds))
    
